""" Test the CLI for run mip-dna """
import logging
import os
from pathlib import Path

from cg.cli.workflow.mip_dna.base import run


def test_cg_dry_run(cli_runner, caplog, case_id, email_adress, dna_mip_context):
    """Test print the MIP run to console"""

    caplog.set_level(logging.INFO)

    pedigree_path = dna_mip_context.get("dna_api").get_pedigree_config_path(case_id=case_id)
    Path.mkdir(pedigree_path.parent, parents=True, exist_ok=True)
    os.link("tests/fixtures/apps/mip/dna/store/pedigree.yaml", pedigree_path)

    # GIVEN a cli function
    # WHEN we run a case in dry run mode
    result = cli_runner.invoke(
        run, ["--dry-run", "--email", email_adress, case_id], obj=dna_mip_context
    )

    # THEN command is run successfully
    assert result.exit_code == 0

    # THEN the command should be printed
    assert (
        f"analyse rd_dna {case_id} --config config.yaml --slurm_quality_of_service normal --email {email_adress}"
        in caplog.text
    )



""" Test cg.cli.upload module """
from datetime import datetime, timedelta

from cg.cli.upload.utils import LinkHelper
from cg.store import Store


def test_all_samples_are_non_tumor(analysis_store, case_id):
    """Test that all samples are non tumor"""

    case_obj = analysis_store.family(case_id)
    assert LinkHelper.all_samples_are_non_tumour(case_obj.links)


def test_all_samples_list_analyses(analysis_store, case_id):
    """Test that all samples have an analysis type"""

    # GIVEN case obj where each sample is wgs analysis
    case_obj = analysis_store.family(case_id)

    # WHEN looking up the analysis type for the samples in the case
    analysis_types = LinkHelper.get_analysis_type_for_each_link(case_obj.links)

    # THEN all the samples should have analysis type 'wgs'
    assert len(set(analysis_types)) == 1 and analysis_types[0] == "wgs"


def test_upload_fails_hard_on_faulty_family(invoke_cli, disk_store: Store):
    """Test that an upload for a missing case does fail hard """

    # GIVEN empty database
    case_id = "dummy_case"

    # WHEN trying to upload with a case that doesn't exist
    result = invoke_cli(["--database", disk_store.uri, "upload", "-f", case_id])

    # THEN it fails hard and reports that it is missing the case
    assert result.exit_code != 0
    assert case_id in result.output
    assert "family not found" in result.output


def test_upload_fails_hard_on_faulty_analysis(invoke_cli, disk_store: Store, helpers):
    """Test that an upload for a missing analysis does fail hard """

    # GIVEN empty database
    case_id = helpers.add_case(disk_store).internal_id

    # WHEN trying to upload with a analysis that doesn't exist
    result = invoke_cli(["--database", disk_store.uri, "upload", "-f", case_id])

    # THEN it fails hard and reports that it is missing the analysis
    assert result.exit_code != 0
    assert case_id in result.output
    assert "no analysis exists" in result.output


def test_upload_doesnt_invoke_dually_for_same_case(invoke_cli, disk_store: Store, helpers):
    """Test that a case that is already uploading can't be uploaded at the same time"""

    # GIVEN an analysis that is already uploading
    case = helpers.add_case(disk_store)
    case_id = case.internal_id
    helpers.add_analysis(disk_store, case=case, uploading=True)

    # WHEN trying to upload it again
    result = invoke_cli(["--database", disk_store.uri, "upload", "-f", case_id])

    # THEN it gracefully skips and reports that it is already uploading
    assert result.exit_code == 0
    assert not result.exception
    assert "already started" in result.output


def test_upload_started_long_time_ago_raises_exception(invoke_cli, disk_store: Store, helpers):
    """Test that an upload for a missing case does fail hard """

    # GIVEN an analysis that is already uploading since a week ago
    case = helpers.add_case(disk_store)
    case_id = case.internal_id
    today = datetime.now()
    upload_started = today - timedelta(hours=100)
    helpers.add_analysis(disk_store, case=case, uploading=True, upload_started=upload_started)

    # WHEN trying to upload an analysis that was started a long time ago
    result = invoke_cli(["--database", disk_store.uri, "upload", "-f", case_id])

    # THEN an exception should have be thrown
    assert result.exit_code != 0
    assert result.exception


def test_upload_force_restart(invoke_cli, disk_store: Store, helpers):
    """Test that a case that is already uploading can be force restarted"""

    # GIVEN an analysis that is already uploading
    case = helpers.add_case(disk_store)
    case_id = case.internal_id

    helpers.add_analysis(disk_store, case=case, uploading=True)

    # WHEN trying to upload it again with the force restart flag
    result = invoke_cli(["--database", disk_store.uri, "upload", "-f", case_id, "-r"])

    # THEN it tries to restart the upload
    assert "already started" not in result.output


def test_upload_uploaded_fails_hard(invoke_cli, disk_store: Store, helpers):
    """Test that a case that has already been uploaded can't be uploaded"""

    # GIVEN an analysis that has been uploaded
    case = helpers.add_case(disk_store)
    case_id = case.internal_id

    helpers.add_analysis(disk_store, case=case, uploaded_at=datetime.now())

    # WHEN trying to upload it again
    result = invoke_cli(["--database", disk_store.uri, "upload", "-f", case_id])

    # THEN it fails hard and reports that it is already uploaded
    assert result.exit_code != 0
    assert "already uploaded" in result.output



""" Add CLI support to start MIP rare disease RNA"""

import logging
from typing import List

import click

from cg.apps.environ import environ_email
from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.apps.lims import LimsAPI
from cg.apps.scout.scoutapi import ScoutAPI
from cg.apps.tb import TrailblazerAPI
from cg.cli.workflow.mip.store import store as store_cmd
from cg.constants import Pipeline
from cg.meta.workflow.mip import MipAnalysisAPI
from cg.store import Store, models
from cg.store.utils import case_exists

LOG = logging.getLogger(__name__)

ARGUMENT_CASE_ID = click.argument("case_id", required=True, type=str)


@click.group("mip-rna")
@click.pass_context
def mip_rna(context: click.Context):
    """Rare disease RNA workflow"""
    context.obj["housekeeper_api"] = HousekeeperAPI(context.obj)
    context.obj["trailblazer_api"] = TrailblazerAPI(context.obj)
    context.obj["scout_api"] = ScoutAPI(context.obj)
    context.obj["lims_api"] = LimsAPI(context.obj)
    context.obj["status_db"] = Store(context.obj["database"])

    context.obj["rna_api"] = MipAnalysisAPI(
        db=context.obj["status_db"],
        hk_api=context.obj["housekeeper_api"],
        tb_api=context.obj["trailblazer_api"],
        scout_api=context.obj["scout_api"],
        lims_api=context.obj["lims_api"],
        script=context.obj["mip-rd-rna"]["script"],
        pipeline=context.obj["mip-rd-rna"]["pipeline"],
        conda_env=context.obj["mip-rd-rna"]["conda_env"],
        root=context.obj["mip-rd-rna"]["root"],
    )


@mip_rna.command()
@ARGUMENT_CASE_ID
@click.pass_context
def link(context: click.Context, case_id: str):
    """Link FASTQ files for all samples in a case"""
    rna_api: MipAnalysisAPI = context.obj["rna_api"]
    case_obj: models.Family = rna_api.db.family(case_id)
    link_objs: List[models.FamilySample] = case_obj.links

    for link_obj in link_objs:
        LOG.info(
            "%s: %s link FASTQ files", link_obj.sample.internal_id, link_obj.family.data_analysis
        )
        if not link_obj.family.data_analysis:
            LOG.warning(
                f"No analysis set for {link_obj.sample.internal_id}, file will still be linked"
            )
            rna_api.link_sample(sample=link_obj.sample, case_id=link_obj.family.internal_id)
        if link_obj.family.data_analysis == str(Pipeline.MIP_RNA):
            rna_api.link_sample(sample=link_obj.sample, case_id=link_obj.family.internal_id)


@mip_rna.command()
@click.option("-d", "--dry-run", "dry_run", is_flag=True, help="print command to console")
@click.option("-e", "--email", help="email to send errors to")
@click.option("--mip-dry-run", "mip_dry_run", is_flag=True, help="Run MIP in dry-run mode")
@click.option("-p", "--priority", type=click.Choice(["low", "normal", "high"]))
@click.option("-sw", "--start-with", help="start mip from this program.")
@ARGUMENT_CASE_ID
@click.pass_context
def run(
    context: click.Context,
    case_id: str,
    dry_run: bool = False,
    mip_dry_run: bool = False,
    priority: str = None,
    email: str = None,
    start_with: str = None,
):
    """Run the analysis for a case"""
    rna_api = context.obj["rna_api"]
    case_obj = rna_api.db.family(case_id)

    if not case_exists(case_obj, case_id):
        raise click.Abort()
    if rna_api.get_analyses_from_trailblazer(case_id=case_obj.internal_id, temp=True):
        LOG.warning("%s: analysis already running", case_obj.internal_id)
        return

    email = email or environ_email()
    kwargs = dict(
        config=context.obj["mip-rd-rna"]["mip_config"],
        case=case_id,
        priority=priority,
        email=email,
        dryrun=mip_dry_run,
        start_with=start_with,
    )
    if dry_run:
        rna_api.run_command(dry_run=dry_run, **kwargs)
        return

    rna_api.run_command(**kwargs)

    if mip_dry_run:
        LOG.info("Executed MIP in dry-run mode - skipping Trailblazer step")
        return

    rna_api.mark_analyses_deleted(case_id=case_id)
    rna_api.add_pending_analysis(
        case_id=case_id,
        email=email,
        type=rna_api.get_application_type(case_id),
        out_dir=rna_api.get_case_output_path(case_id).as_posix(),
        config_path=rna_api.get_slurm_job_ids_path(case_id).as_posix(),
        priority="normal",
        data_analysis=Pipeline.MIP_RNA,
    )
    rna_api.set_statusdb_action(case_id=case_id, action="running")
    LOG.info("MIP rd-rna run started!")


@mip_rna.command("config-case")
@click.option("-d", "--dry-run", "dry_run", is_flag=True, help="Print pedigree config to console")
@ARGUMENT_CASE_ID
@click.pass_context
def config_case(context: click.Context, case_id: str, dry_run: bool = False):
    """Generate a pedigree config for the case"""
    rna_api = context.obj["rna_api"]

    case_obj = rna_api.db.family(case_id)
    if not case_exists(case_obj, case_id):
        LOG.error("Case %s does not exist!", case_id)
        context.abort()
    config_data = rna_api.pedigree_config(case_obj, pipeline=Pipeline.MIP_RNA)
    if dry_run:
        click.echo(config_data)
        return
    out_path = rna_api.write_pedigree_config(
        data=config_data,
        out_dir=rna_api.get_case_output_path(case_id),
        pedigree_config_path=rna_api.get_pedigree_config_path(case_id),
    )
    LOG.info(f"Config saved to: {out_path}")


mip_rna.add_command(store_cmd)



"""Fixtures for testing the madeline cg app"""

from pathlib import Path
from typing import Dict, List, Optional

import pytest
from cg.apps.madeline.api import MadelineAPI
from tests.mocks.process_mock import ProcessMock


@pytest.fixture(name="madeline_output")
def fixture_madeline_output(apps_dir: Path) -> Path:
    """Path to madeline output"""
    return apps_dir / "madeline" / "madeline.xml"


@pytest.fixture(name="madeline_columns")
def fixture_madeline_columns() -> Dict[str, str]:
    """return a dictionary with madeline columns"""
    columns = {
        "case": "FamilyId",
        "sample": "IndividualId",
        "sex": "Gender",
        "father": "Father",
        "mother": "Mother",
        "deceased": "Deceased",
        "proband": "Proband",
        "status": "Affected",
    }
    return columns


@pytest.fixture(name="mother")
def fixture_mother() -> Dict[str, Optional[str]]:
    """return a dictionary with ind info"""
    ind_info = {
        "sample": "mother",
        "sex": "female",
        "proband": False,
        "status": "unaffected",
    }
    return ind_info


@pytest.fixture(name="father")
def fixture_father() -> Dict[str, Optional[str]]:
    """return a dictionary with ind info"""
    ind_info = {
        "sample": "father",
        "sex": "male",
        "proband": False,
        "status": "unaffected",
    }
    return ind_info


@pytest.fixture(name="proband")
def fixture_proband() -> Dict[str, Optional[str]]:
    """return a dictionary with ind info"""
    ind_info = {
        "sample": "proband",
        "sex": "female",
        "father": "father",
        "mother": "mother",
        "deceased": False,
        "proband": True,
        "status": "affected",
    }
    return ind_info


@pytest.fixture(name="trio")
def fixture_trio(proband: dict, mother: dict, father: dict) -> List[Dict[str, Optional[str]]]:
    """return a list with a trio"""
    return [proband, mother, father]


@pytest.fixture
def madeline_input(proband: dict) -> List[str]:
    """return a iterable with madeline formated lines"""
    individuals = [proband]
    case_id = "test"
    _input = []
    for line in MadelineAPI.make_ped(case_id, individuals):
        _input.append(line)

    return _input


@pytest.fixture(name="madeline_api")
def fixture_madeline_api() -> MadelineAPI:
    """Return a madeline API with mocked process"""
    binary_path = "madeline"
    config = {"madeline_exe": binary_path}
    madeline_api: MadelineAPI = MadelineAPI(config)
    madeline_process: ProcessMock = ProcessMock(binary=binary_path)
    madeline_api.process = madeline_process
    return madeline_api


@pytest.fixture(name="populated_madeline_api")
def fixture_populated_madeline_api(madeline_output: Path, madeline_api: MadelineAPI) -> MadelineAPI:
    """Return a madeline API populated with some output"""
    with open(madeline_output, "r") as output:
        madeline_content = output.read()
    madeline_api.process.set_stdout(madeline_content)
    return madeline_api



"""CLI for deleting records in statusDB """

import logging

import click
from cg.cli.delete.case import case
from cg.cli.delete.cases import cases
from cg.store import Store

LOG = logging.getLogger(__name__)


@click.group()
@click.pass_context
def delete(context):
    """delete database records in CG."""
    LOG.info("Running CG delete")
    context.obj["status_db"] = Store(context.obj["database"])


delete.add_command(case)
delete.add_command(cases)



"""
    Conftest file for pytest fixtures that needs to be shared for multiple tests
"""
import copy
import datetime as dt
import json
import logging
import os
import shutil
from pathlib import Path

import pytest
import ruamel.yaml
from cg.apps.gt import GenotypeAPI
from cg.apps.hermes.hermes_api import HermesApi
from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.apps.mip import parse_qcmetrics, parse_sampleinfo
from cg.constants import Pipeline
from cg.models import CompressionData
from cg.store import Store

from .mocks.crunchy import MockCrunchyAPI
from .mocks.hk_mock import MockHousekeeperAPI
from .mocks.limsmock import MockLimsAPI
from .mocks.madeline import MockMadelineAPI
from .mocks.osticket import MockOsTicket
from .mocks.process_mock import ProcessMock
from .mocks.scout import MockScoutAPI
from .mocks.tb_mock import MockTB
from .small_helpers import SmallHelpers
from .store_helpers import StoreHelpers

CHANJO_CONFIG = {"chanjo": {"config_path": "chanjo_config", "binary_path": "chanjo"}}
CRUNCHY_CONFIG = {
    "crunchy": {
        "cram_reference": "/path/to/fasta",
        "slurm": {"account": "mock_account", "mail_user": "mock_mail", "conda_env": "mock_env"},
    }
}


LOG = logging.getLogger(__name__)

# Case fixtures


@pytest.fixture(name="user_name")
def fixture_user_name() -> str:
    return "Paul Anderson"


@pytest.fixture(name="user_mail")
def fixture_user_mail() -> str:
    return "paul@magnolia.com"


@pytest.fixture(name="email_adress")
def fixture_email_adress() -> str:
    return "james.holden@scilifelab.se"


@pytest.fixture(name="case_id")
def fixture_case_id() -> str:
    """Return a case id"""
    return "yellowhog"


@pytest.fixture(name="sample_id")
def fixture_sample_id() -> str:
    """ Returns a sample id """
    return "ADM1"


@pytest.fixture(name="family_name")
def fixture_family_name() -> str:
    """Return a case name"""
    return "case"


@pytest.fixture(name="customer_id")
def fixture_customer_id() -> str:
    """Return a customer id"""
    return "cust000"


@pytest.fixture(name="ticket_nr")
def fixture_ticket_nr() -> int:
    """Return a ticket nr"""
    return 123456


@pytest.fixture(scope="function", name="analysis_family_single_case")
def fixture_analysis_family_single(case_id: str, family_name: str, ticket_nr: int) -> dict:
    """Build an example case."""
    return {
        "name": family_name,
        "internal_id": case_id,
        "data_analysis": str(Pipeline.MIP_DNA),
        "application_type": "wgs",
        "panels": ["IEM", "EP"],
        "samples": [
            {
                "name": "proband",
                "sex": "male",
                "internal_id": "ADM1",
                "status": "affected",
                "ticket_number": ticket_nr,
                "reads": 5000000000,
                "capture_kit": "GMSmyeloid",
            }
        ],
    }


@pytest.fixture(scope="function", name="analysis_family")
def fixture_analysis_family(case_id: str, family_name: str, ticket_nr: int) -> dict:
    """Return a dictionary with information from a analysis case"""
    return {
        "name": family_name,
        "internal_id": case_id,
        "data_analysis": str(Pipeline.MIP_DNA),
        "application_type": "wgs",
        "panels": ["IEM", "EP"],
        "samples": [
            {
                "name": "child",
                "sex": "male",
                "internal_id": "ADM1",
                "father": "ADM2",
                "mother": "ADM3",
                "status": "affected",
                "ticket_number": ticket_nr,
                "reads": 5000000,
                "capture_kit": "GMSmyeloid",
            },
            {
                "name": "father",
                "sex": "male",
                "internal_id": "ADM2",
                "status": "unaffected",
                "ticket_number": ticket_nr,
                "reads": 6000000,
                "capture_kit": "GMSmyeloid",
            },
            {
                "name": "mother",
                "sex": "female",
                "internal_id": "ADM3",
                "status": "unaffected",
                "ticket_number": ticket_nr,
                "reads": 7000000,
                "capture_kit": "GMSmyeloid",
            },
        ],
    }


# Config fixtures


@pytest.fixture
def chanjo_config_dict():
    """Chanjo configs"""
    _config = {}
    _config.update(CHANJO_CONFIG)
    return _config


@pytest.fixture
def crunchy_config_dict():
    """Crunchy configs"""
    _config = {}
    _config.update(CRUNCHY_CONFIG)
    return _config


@pytest.fixture(name="hk_config_dict")
def fixture_hk_config_dict(root_path):
    """Crunchy configs"""
    return {
        "housekeeper": {
            "database": "sqlite:///:memory:",
            "root": str(root_path),
        }
    }


@pytest.fixture(name="genotype_config")
def fixture_genotype_config() -> dict:
    """
    genotype config fixture
    """
    return {
        "genotype": {
            "database": "database",
            "config_path": "config/path",
            "binary_path": "gtdb",
        }
    }


# Api fixtures


@pytest.fixture(name="genotype_api")
def fixture_genotype_api(genotype_config: dict) -> GenotypeAPI:
    """
    genotype API fixture
    """
    _genotype_api = GenotypeAPI(genotype_config)
    _genotype_api.set_dry_run(True)
    return _genotype_api


@pytest.fixture(scope="function")
def madeline_api(madeline_output) -> MockMadelineAPI:
    """madeline_api fixture"""
    _api = MockMadelineAPI()
    _api.set_outpath(madeline_output)

    return _api


@pytest.fixture(name="ticket_number")
def fixture_ticket_number() -> int:
    """Return a ticket number for testing"""
    return 123456


@pytest.fixture(name="osticket")
def fixture_os_ticket(ticket_number: int) -> MockOsTicket:
    """Return a api that mock the os ticket api"""
    api = MockOsTicket()
    api.set_ticket_nr(ticket_number)
    return api


# Files fixtures

# Common file fixtures
@pytest.fixture(name="fixtures_dir")
def fixture_fixtures_dir() -> Path:
    """Return the path to the fixtures dir"""
    return Path("tests/fixtures")


@pytest.fixture(name="analysis_dir")
def fixture_analysis_dir(fixtures_dir: Path) -> Path:
    """Return the path to the analysis dir"""
    return fixtures_dir / "analysis"


@pytest.fixture(name="apps_dir")
def fixture_apps_dir(fixtures_dir: Path) -> Path:
    """Return the path to the apps dir"""
    return fixtures_dir / "apps"


@pytest.fixture(name="fastq_dir")
def fixture_fastq_dir(fixtures_dir: Path) -> Path:
    """Return the path to the fastq files dir"""
    return fixtures_dir / "fastq"


@pytest.fixture(scope="function", name="project_dir")
def fixture_project_dir(tmpdir_factory):
    """Path to a temporary directory where intermediate files can be stored"""
    my_tmpdir = Path(tmpdir_factory.mktemp("data"))
    yield my_tmpdir
    shutil.rmtree(str(my_tmpdir))


@pytest.fixture(scope="function")
def tmp_file(project_dir):
    """Get a temp file"""
    return project_dir / "test"


@pytest.fixture(name="orderforms")
def fixture_orderform(fixtures_dir: Path) -> Path:
    """Return the path to the directory with orderforms"""
    return fixtures_dir / "orderforms"


@pytest.fixture(name="mip_dna_store_files")
def fixture_mip_dna_store_files(apps_dir: Path) -> Path:
    """Return the path to the directory with mip dna store files"""
    return apps_dir / "mip" / "dna" / "store"


@pytest.fixture(name="mip_analysis_dir")
def fixture_mip_analysis_dir(analysis_dir: Path) -> Path:
    """Return the path to the directory with mip analysis files"""
    return analysis_dir / "mip"


@pytest.fixture(name="balsamic_analysis_dir")
def fixture_balsamic_analysis_dir(analysis_dir: Path) -> Path:
    """Return the path to the directory with balsamic analysis files"""
    return analysis_dir / "balsamic"


@pytest.fixture(name="balsamic_panel_analysis_dir")
def fixture_balsamic_panel_analysis_dir(balsamic_analysis_dir: Path) -> Path:
    """Return the path to the directory with balsamic analysis files"""
    return balsamic_analysis_dir / "tn_panel"


@pytest.fixture(name="mip_dna_analysis_dir")
def fixture_mip_dna_analysis_dir(mip_analysis_dir: Path) -> Path:
    """Return the path to the directory with mip dna analysis files"""
    return mip_analysis_dir / "dna"


@pytest.fixture(name="sample1_cram")
def fixture_sample1_cram(mip_dna_analysis_dir: Path) -> Path:
    """Return the path to the cram file for sample 1"""
    return mip_dna_analysis_dir / "adm1.cram"


@pytest.fixture(name="mip_deliverables_file")
def fixture_mip_deliverables_files(mip_dna_store_files: Path) -> Path:
    """Fixture for general deliverables file in mip"""
    return mip_dna_store_files / "case_id_deliverables.yaml"


@pytest.fixture(name="vcf_file")
def fixture_vcf_file(mip_dna_store_files: Path) -> Path:
    """Return the path to to a vcf file"""
    return mip_dna_store_files / "yellowhog_clinical_selected.vcf"


# Orderform fixtures


@pytest.fixture
def microbial_orderform(orderforms: Path) -> str:
    """Orderform fixture for microbial samples"""
    return Path(orderforms / "1603.9.microbial.xlsx").as_posix()


@pytest.fixture
def rml_orderform(orderforms: Path) -> str:
    """Orderform fixture for RML samples"""
    return Path(orderforms / "1604.10.rml.xlsx").as_posix()


@pytest.fixture(name="madeline_output")
def fixture_madeline_output(apps_dir: Path) -> str:
    """File with madeline output"""
    _file = apps_dir / "madeline/madeline.xml"
    return str(_file)


@pytest.fixture
def mip_order_to_submit() -> dict:
    """Load an example scout order."""
    return json.load(open("tests/fixtures/orders/mip.json"))


@pytest.fixture
def mip_json_order_to_submit() -> dict:
    """Load an example json scout order."""
    return json.load(open("tests/fixtures/orders/mip-json.json"))


@pytest.fixture
def mip_rna_order_to_submit() -> dict:
    """Load an example rna order."""
    return json.load(open("tests/fixtures/orders/mip_rna.json"))


@pytest.fixture
def external_order_to_submit() -> dict:
    """Load an example external order."""
    return json.load(open("tests/fixtures/orders/external.json"))


@pytest.fixture
def fastq_order_to_submit() -> dict:
    """Load an example fastq order."""
    return json.load(open("tests/fixtures/orders/fastq.json"))


@pytest.fixture
def rml_order_to_submit() -> dict:
    """Load an example rml order."""
    return json.load(open("tests/fixtures/orders/rml.json"))


@pytest.fixture
def metagenome_order_to_submit() -> dict:
    """Load an example metagenome order."""
    return json.load(open("tests/fixtures/orders/metagenome.json"))


@pytest.fixture
def microbial_order_to_submit() -> dict:
    """Load an example microbial order."""
    return json.load(open("tests/fixtures/orders/microsalt.json"))


@pytest.fixture
def balsamic_order_to_submit() -> dict:
    """Load an example cancer order."""
    return json.load(open("tests/fixtures/orders/balsamic.json"))


# Compression fixtures


@pytest.fixture(scope="function", name="run_name")
def fixture_run_name() -> str:
    """Return the name of a fastq run"""
    return "fastq_run"


@pytest.fixture(scope="function", name="original_fastq_data")
def fixture_original_fastq_data(fastq_dir: Path, run_name) -> CompressionData:
    """Return a compression object with a path to the original fastq files"""

    return CompressionData(fastq_dir / run_name)


@pytest.fixture(scope="function", name="fastq_stub")
def fixture_fastq_stub(project_dir: Path, run_name: str) -> Path:
    """Creates a path to the base format of a fastq run"""
    return project_dir / run_name


@pytest.fixture(scope="function", name="compression_object")
def fixture_compression_object(
    fastq_stub: Path, original_fastq_data: CompressionData
) -> CompressionData:
    """Creates compression data object with information about files used in fastq compression"""
    working_files = CompressionData(fastq_stub)
    shutil.copy(str(original_fastq_data.fastq_first), str(working_files.fastq_first))
    shutil.copy(str(original_fastq_data.fastq_second), str(working_files.fastq_second))
    return working_files


# Unknown file fixtures


@pytest.fixture(name="case_qc_metrics")
def fixture_case_qc_metrics(apps_dir: Path) -> Path:
    """Return the path to a qc metrics file with case data"""
    return Path("tests/fixtures/apps/mip/case_qc_metrics.yaml")


@pytest.fixture(name="bcf_file")
def fixture_bcf_file(apps_dir: Path) -> Path:
    """Return the path to a bcf file"""
    return apps_dir / "gt" / "yellowhog.bcf"


@pytest.fixture(scope="session", name="files")
def fixture_files():
    """Trailblazer api for mip files"""
    return {
        "config": "tests/fixtures/apps/mip/dna/store/case_config.yaml",
        "sampleinfo": "tests/fixtures/apps/mip/dna/store/case_qc_sample_info.yaml",
        "qcmetrics": "tests/fixtures/apps/mip/case_qc_metrics.yaml",
        "rna_config": "tests/fixtures/apps/mip/rna/case_config.yaml",
        "rna_sampleinfo": "tests/fixtures/apps/mip/rna/case_qc_sampleinfo.yaml",
        "rna_config_store": "tests/fixtures/apps/mip/rna/store/case_config.yaml",
        "rna_sampleinfo_store": "tests/fixtures/apps/mip/rna/store/case_qc_sample_info.yaml",
        "mip_rna_deliverables": "test/fixtures/apps/mip/rna/store/case_deliverables.yaml",
        "dna_config_store": "tests/fixtures/apps/mip/dna/store/case_config.yaml",
        "dna_sampleinfo_store": "tests/fixtures/apps/mip/dna/store/case_qc_sample_info.yaml",
        "mip_dna_deliverables": "test/fixtures/apps/mip/dna/store/case_deliverables.yaml",
    }


@pytest.fixture(scope="function", name="bed_file")
def fixture_bed_file(analysis_dir) -> str:
    """Get the path to a bed file file"""
    return str(analysis_dir / "sample_coverage.bed")


@pytest.fixture(scope="session", name="files_raw")
def fixture_files_raw(files):
    """Get some raw files"""
    return {
        "config": ruamel.yaml.safe_load(open(files["config"])),
        "sampleinfo": ruamel.yaml.safe_load(open(files["sampleinfo"])),
        "qcmetrics": ruamel.yaml.safe_load(open(files["qcmetrics"])),
        "rna_config": ruamel.yaml.safe_load(open(files["rna_config"])),
        "rna_sampleinfo": ruamel.yaml.safe_load(open(files["rna_sampleinfo"])),
        "rna_config_store": ruamel.yaml.safe_load(open(files["rna_config_store"])),
        "rna_sampleinfo_store": ruamel.yaml.safe_load(open(files["rna_sampleinfo_store"])),
        "dna_config_store": ruamel.yaml.safe_load(open(files["dna_config_store"])),
        "dna_sampleinfo_store": ruamel.yaml.safe_load(open(files["dna_sampleinfo_store"])),
    }


@pytest.fixture(scope="session")
def files_data(files_raw):
    """Get some data files"""
    return {
        "config": parse_sampleinfo.parse_config(files_raw["config"]),
        "sampleinfo": parse_sampleinfo.parse_sampleinfo(files_raw["sampleinfo"]),
        "qcmetrics": parse_qcmetrics.parse_qcmetrics(files_raw["qcmetrics"]),
        "rna_config": parse_sampleinfo.parse_config(files_raw["rna_config"]),
        "rna_sampleinfo": parse_sampleinfo.parse_sampleinfo_rna(files_raw["rna_sampleinfo"]),
        "rna_config_store": parse_sampleinfo.parse_config(files_raw["rna_config_store"]),
        "rna_sampleinfo_store": parse_sampleinfo.parse_sampleinfo(
            files_raw["rna_sampleinfo_store"]
        ),
        "dna_config_store": parse_sampleinfo.parse_config(files_raw["dna_config_store"]),
        "dna_sampleinfo_store": parse_sampleinfo.parse_sampleinfo(
            files_raw["dna_sampleinfo_store"]
        ),
    }


# Helper fixtures


@pytest.fixture(name="helpers")
def fixture_helpers():
    """Return a class with helper functions for the stores"""
    return StoreHelpers()


@pytest.fixture(name="small_helpers")
def fixture_small_helpers():
    """Return a class with small helper functions"""
    return SmallHelpers()


# HK fixtures


@pytest.fixture(name="root_path")
def fixture_root_path(project_dir: Path) -> Path:
    """Return the path to a hk bundles dir"""
    _root_path = project_dir / "bundles"
    _root_path.mkdir(parents=True, exist_ok=True)
    return _root_path


@pytest.fixture(scope="function", name="timestamp")
def fixture_timestamp() -> dt.datetime:
    """Return a time stamp in date time format"""
    return dt.datetime(2020, 5, 1)


@pytest.fixture(scope="function", name="later_timestamp")
def fixture_later_timestamp() -> dt.datetime:
    """Return a time stamp in date time format"""
    return dt.datetime(2020, 6, 1)


@pytest.fixture(scope="function", name="timestamp_today")
def fixture_timestamp_today() -> dt.datetime:
    """Return a time stamp of todays date in date time format"""
    return dt.datetime.now()


@pytest.fixture(scope="function", name="timestamp_yesterday")
def fixture_timestamp_yesterday(timestamp_today: dt.datetime) -> dt.datetime:
    """Return a time stamp of yesterdays date in date time format"""
    return timestamp_today - dt.timedelta(days=1)


@pytest.fixture(scope="function", name="hk_bundle_data")
def fixture_hk_bundle_data(case_id: str, bed_file: str, timestamp: dt.datetime) -> dict:
    """Get some bundle data for housekeeper"""
    return {
        "name": case_id,
        "created": timestamp,
        "expires": timestamp,
        "files": [{"path": bed_file, "archive": False, "tags": ["bed", "sample"]}],
    }


@pytest.fixture(scope="function", name="sample_hk_bundle_no_files")
def fixture_sample_hk_bundle_no_files(sample: str, timestamp: dt.datetime) -> dict:
    """Create a complete bundle mock for testing compression"""
    return {
        "name": sample,
        "created": timestamp,
        "expires": timestamp,
        "files": [],
    }


@pytest.fixture(scope="function", name="case_hk_bundle_no_files")
def fixture_case_hk_bundle_no_files(case_id: str, timestamp: dt.datetime) -> dict:
    """Create a complete bundle mock for testing compression"""
    return {
        "name": case_id,
        "created": timestamp,
        "expires": timestamp,
        "files": [],
    }


@pytest.fixture(scope="function", name="compress_hk_fastq_bundle")
def fixture_compress_hk_fastq_bundle(
    compression_object: CompressionData, sample_hk_bundle_no_files: dict
) -> dict:
    """Create a complete bundle mock for testing compression

    This bundle contains a pair of fastq files.
    """
    hk_bundle_data = copy.deepcopy(sample_hk_bundle_no_files)

    first_fastq = compression_object.fastq_first
    second_fastq = compression_object.fastq_second
    for fastq_file in [first_fastq, second_fastq]:
        fastq_file.touch()
        # We need to set the time to an old date
        # Create a older date
        # Convert the date to a float
        before_timestamp = dt.datetime.timestamp(dt.datetime(2020, 1, 1))
        # Update the utime so file looks old
        os.utime(fastq_file, (before_timestamp, before_timestamp))
        fastq_file_info = {"path": str(fastq_file), "archive": False, "tags": ["fastq"]}

        hk_bundle_data["files"].append(fastq_file_info)

    return hk_bundle_data


@pytest.fixture(name="housekeeper_api")
def fixture_housekeeper_api(hk_config_dict: dict) -> MockHousekeeperAPI:
    """Setup Housekeeper store."""
    return MockHousekeeperAPI(hk_config_dict)


@pytest.fixture(scope="function", name="real_housekeeper_api")
def fixture_real_housekeeper_api(hk_config_dict: dict) -> HousekeeperAPI:
    """Setup a real Housekeeper store."""
    _api = HousekeeperAPI(hk_config_dict)
    _api.initialise_db()
    yield _api


@pytest.fixture(scope="function", name="populated_housekeeper_api")
def fixture_populated_housekeeper_api(
    housekeeper_api: MockHousekeeperAPI, hk_bundle_data: dict, helpers
) -> MockHousekeeperAPI:
    """Setup a Housekeeper store with some data."""
    hk_api = housekeeper_api
    helpers.ensure_hk_bundle(hk_api, hk_bundle_data)
    return hk_api


@pytest.fixture(scope="function", name="hk_version_obj")
def fixture_hk_version_obj(
    housekeeper_api: MockHousekeeperAPI, hk_bundle_data: dict, helpers
) -> MockHousekeeperAPI:
    """Get a housekeeper version object"""
    return helpers.ensure_hk_version(housekeeper_api, hk_bundle_data)


# Process Mock


@pytest.fixture(name="process")
def fixture_process() -> ProcessMock:
    """Returns a mocked process"""
    return ProcessMock()


# Hermes mock


@pytest.fixture(name="hermes_process")
def fixture_hermes_process() -> ProcessMock:
    """Return a mocked hermes process"""
    return ProcessMock(binary="hermes")


@pytest.fixture(name="hermes_api")
def fixture_hermes_api(hermes_process: ProcessMock) -> HermesApi:
    """Return a hermes api with a mocked process"""
    hermes_config = {"hermes": {"deploy_config": "deploy_config", "binary_path": "/bin/true"}}
    hermes_api = HermesApi(config=hermes_config)
    hermes_api.process = hermes_process
    return hermes_api


# Scout fixtures


@pytest.fixture(scope="function", name="scout_api")
def fixture_scout_api() -> MockScoutAPI:
    """Setup Scout api."""
    return MockScoutAPI()


# Crunchy fixtures


@pytest.fixture(scope="function", name="crunchy_api")
def fixture_crunchy_api():
    """Setup Crunchy api."""
    return MockCrunchyAPI()


# Store fixtures


@pytest.fixture(scope="function", name="analysis_store")
def fixture_analysis_store(
    base_store: Store, analysis_family: dict, wgs_application_tag: str, helpers
):
    """Setup a store instance for testing analysis API."""
    helpers.ensure_case_from_dict(
        base_store, case_info=analysis_family, app_tag=wgs_application_tag
    )

    yield base_store


@pytest.fixture(scope="function", name="analysis_store_trio")
def fixture_analysis_store_trio(analysis_store):
    """Setup a store instance with a trion loaded for testing analysis API."""

    yield analysis_store


@pytest.fixture(scope="function", name="analysis_store_single_case")
def fixture_analysis_store_single(base_store, analysis_family_single_case, helpers):
    """Setup a store instance with a single ind case for testing analysis API."""
    helpers.ensure_case_from_dict(base_store, case_info=analysis_family_single_case)

    yield base_store


@pytest.fixture(scope="function", name="customer_group")
def fixture_customer_group() -> str:
    """Return a default customer group"""
    return "all_customers"


@pytest.fixture(scope="function", name="customer_production")
def fixture_customer_production(customer_group) -> dict:
    """Return a dictionary with information about the prod customer"""
    return dict(
        customer_id="cust000",
        name="Production",
        scout_access=True,
        customer_group=customer_group,
    )


@pytest.fixture(scope="function", name="external_wgs_application_tag")
def fixture_external_wgs_application_tag() -> str:
    """Return the external wgs app tag"""
    return "WGXCUSC000"


@pytest.fixture(scope="function", name="external_wgs_info")
def fixture_external_wgs_info(external_wgs_application_tag) -> dict:
    """Return a dictionary with information external WGS application"""
    return dict(
        application_tag=external_wgs_application_tag,
        application_type="wgs",
        description="External WGS",
        is_external=True,
        target_reads=10,
    )


@pytest.fixture(scope="function", name="external_wes_application_tag")
def fixture_external_wes_application_tag() -> str:
    """Return the external whole exome sequencing app tag"""
    return "EXXCUSR000"


@pytest.fixture(scope="function", name="external_wes_info")
def fixture_external_wes_info(external_wes_application_tag) -> dict:
    """Return a dictionary with information external WES application"""
    return dict(
        application_tag=external_wes_application_tag,
        application_type="wes",
        description="External WES",
        is_external=True,
        target_reads=10,
    )


@pytest.fixture(scope="function", name="wgs_application_tag")
def fixture_wgs_application_tag() -> str:
    """Return the wgs app tag"""
    return "WGSPCFC060"


@pytest.fixture(scope="function", name="wgs_application_info")
def fixture_wgs_application_info(wgs_application_tag) -> dict:
    """Return a dictionary with information the WGS application"""
    return dict(
        application_tag=wgs_application_tag,
        application_type="wgs",
        description="WGS, double",
        sequencing_depth=30,
        is_external=True,
        is_accredited=True,
        target_reads=10,
    )


@pytest.fixture(scope="function", name="store")
def fixture_store() -> Store:
    """Fixture with a CG store"""
    _store = Store(uri="sqlite:///")
    _store.create_all()
    yield _store
    _store.drop_all()


@pytest.fixture(name="apptag_rna")
def fixture_apptag_rna() -> str:
    return "RNAPOAR025"


@pytest.fixture(scope="function", name="base_store")
def fixture_base_store(store: Store, apptag_rna: str) -> Store:
    """Setup and example store."""
    customer_group = store.add_customer_group("all_customers", "all customers")

    store.add_commit(customer_group)
    customers = [
        store.add_customer(
            "cust000",
            "Production",
            scout_access=True,
            customer_group=customer_group,
            invoice_address="Test street",
            invoice_reference="ABCDEF",
        ),
        store.add_customer(
            "cust001",
            "Customer",
            scout_access=False,
            customer_group=customer_group,
            invoice_address="Test street",
            invoice_reference="ABCDEF",
        ),
        store.add_customer(
            "cust002",
            "Karolinska",
            scout_access=True,
            customer_group=customer_group,
            invoice_address="Test street",
            invoice_reference="ABCDEF",
        ),
        store.add_customer(
            "cust003",
            "CMMS",
            scout_access=True,
            customer_group=customer_group,
            invoice_address="Test street",
            invoice_reference="ABCDEF",
        ),
    ]
    store.add_commit(customers)
    applications = [
        store.add_application(
            tag="WGXCUSC000",
            category="wgs",
            description="External WGS",
            sequencing_depth=0,
            is_external=True,
            percent_kth=80,
            percent_reads_guaranteed=75,
            target_reads=10,
        ),
        store.add_application(
            tag="EXXCUSR000",
            category="wes",
            description="External WES",
            sequencing_depth=0,
            is_external=True,
            percent_kth=80,
            percent_reads_guaranteed=75,
            target_reads=10,
        ),
        store.add_application(
            tag="WGSPCFC060",
            category="wgs",
            description="WGS, double",
            sequencing_depth=30,
            accredited=True,
            percent_kth=80,
            percent_reads_guaranteed=75,
            target_reads=10,
        ),
        store.add_application(
            tag="RMLS05R150",
            category="rml",
            description="Ready-made",
            sequencing_depth=0,
            percent_kth=80,
            percent_reads_guaranteed=75,
            target_reads=10,
        ),
        store.add_application(
            tag="WGTPCFC030",
            category="wgs",
            description="WGS trio",
            is_accredited=True,
            sequencing_depth=30,
            target_reads=30,
            limitations="some",
            percent_kth=80,
            percent_reads_guaranteed=75,
        ),
        store.add_application(
            tag="METLIFR020",
            category="wgs",
            description="Whole genome metagenomics",
            sequencing_depth=0,
            target_reads=400000,
            percent_kth=80,
            percent_reads_guaranteed=75,
        ),
        store.add_application(
            tag="METNXTR020",
            category="wgs",
            description="Metagenomics",
            sequencing_depth=0,
            target_reads=200000,
            percent_kth=80,
            percent_reads_guaranteed=75,
        ),
        store.add_application(
            tag="MWRNXTR003",
            category="mic",
            description="Microbial whole genome ",
            sequencing_depth=0,
            percent_kth=80,
            percent_reads_guaranteed=75,
            target_reads=10,
        ),
        store.add_application(
            tag=apptag_rna,
            category="tgs",
            description="RNA seq, poly-A based priming",
            percent_kth=80,
            percent_reads_guaranteed=75,
            sequencing_depth=25,
            accredited=True,
            target_reads=10,
        ),
    ]

    store.add_commit(applications)

    prices = {"standard": 10, "priority": 20, "express": 30, "research": 5}
    versions = [
        store.add_version(application, 1, valid_from=dt.datetime.now(), prices=prices)
        for application in applications
    ]
    store.add_commit(versions)

    beds = [store.add_bed("Bed")]
    store.add_commit(beds)
    bed_versions = [store.add_bed_version(bed, 1, "Bed.bed") for bed in beds]
    store.add_commit(bed_versions)

    organism = store.add_organism("C. jejuni", "C. jejuni")
    store.add_commit(organism)

    yield store


@pytest.fixture(scope="function")
def sample_store(base_store) -> Store:
    """Populate store with samples."""
    new_samples = [
        base_store.add_sample("ordered", sex="male"),
        base_store.add_sample("received", sex="unknown", received=dt.datetime.now()),
        base_store.add_sample(
            "received-prepared",
            sex="unknown",
            received=dt.datetime.now(),
            prepared_at=dt.datetime.now(),
        ),
        base_store.add_sample("external", sex="female", external=True),
        base_store.add_sample(
            "external-received", sex="female", external=True, received=dt.datetime.now()
        ),
        base_store.add_sample(
            "sequenced",
            sex="male",
            received=dt.datetime.now(),
            prepared_at=dt.datetime.now(),
            sequenced_at=dt.datetime.now(),
            reads=(310 * 1000000),
        ),
        base_store.add_sample(
            "sequenced-partly",
            sex="male",
            received=dt.datetime.now(),
            prepared_at=dt.datetime.now(),
            reads=(250 * 1000000),
        ),
    ]
    customer = base_store.customers().first()
    external_app = base_store.application("WGXCUSC000").versions[0]
    wgs_app = base_store.application("WGTPCFC030").versions[0]
    for sample in new_samples:
        sample.customer = customer
        sample.application_version = external_app if "external" in sample.name else wgs_app
    base_store.add_commit(new_samples)
    return base_store


@pytest.fixture(scope="function")
def disk_store(cli_runner, invoke_cli) -> Store:
    """Store on disk"""
    database = "./test_db.sqlite3"
    database_path = Path(database)
    with cli_runner.isolated_filesystem():
        assert database_path.exists() is False

        database_uri = f"sqlite:///{database}"
        # WHEN calling "init"
        result = invoke_cli(["--database", database_uri, "init"])

        # THEN it should setup the database with some tables
        assert result.exit_code == 0
        assert database_path.exists()
        assert len(Store(database_uri).engine.table_names()) > 0

        yield Store(database_uri)


@pytest.fixture(scope="function", name="trailblazer_api")
def fixture_trailblazer_api() -> MockTB:
    return MockTB()


@pytest.fixture(scope="function", name="lims_api")
def fixture_lims_api() -> MockLimsAPI:
    return MockLimsAPI()



"""Module to validate that required delivery report data exists"""
from cg.exc import ValidationError
from cg.meta.report.sample_helper import SampleHelper
from cg.store import Store


REQUIRED_REPORT_FIELDS = [
    "report_version",
    "case",
    "customer_name",
    "today",
    "panels",
    "customer_invoice_address",
    "scout_access",
    "accredited",
    "pipeline",
    "pipeline_version",
    "genome_build",
]

REQUIRED_GENERIC_SAMPLE_FIELDS = [
    "name",
    "internal_id",
    "sex",
    "status",
    "ticket",
    "application",
    "ordered_at",
    "received_at",
    "delivered_at",
    "processing_time",
    "analysis_sex",
]

REQUIRED_ANALYSIS_SAMPLE_FIELDS = [
    "mapped_reads",
    "duplicates",
    "analysis_sex",
    "target_coverage",
    "target_completeness",
]


class ReportValidator:
    """API to validate report data"""

    def __init__(self, store: Store):
        self._sample_helper = SampleHelper(store)
        self._attributes_missing_values = None

    def has_required_data(self, report_data: dict, case_id: str) -> bool:
        """Main method to validate report data"""

        self._attributes_missing_values = []

        self._check_required_general_report_data(report_data)

        for sample in report_data["samples"]:
            self._check_required_sample_attributes(sample, case_id)

        return self.get_missing_attributes() == []

    def get_missing_attributes(self) -> [str]:
        """Returns a list of all attributes that seems to be missing in the report data set"""
        if self._attributes_missing_values is None:
            raise ValidationError(
                "get_missing_attributes() can only be run after " "has_required_data(...)"
            )

        return self._attributes_missing_values

    def _collect_missing_attributes(self, a_dict: dict, keys: list) -> None:
        """Gathers all attributes that should exist but is without a value"""
        for key in keys:
            if a_dict.get(key) is None:
                self._attributes_missing_values.append(key)

    def _check_required_sample_attributes_for_all(self, sample: dict) -> None:
        """Checks attributes that should exist on all samples"""
        self._collect_missing_attributes(sample, REQUIRED_GENERIC_SAMPLE_FIELDS)

    def _check_required_sample_attributes(self, sample: dict, case_id: str) -> None:
        """Checks attributes for a sample"""
        self._check_required_sample_attributes_for_all(sample)

        if not sample.get("internal_id"):
            return

        if self._sample_helper.is_ready_made_sample(sample["internal_id"]):
            raise ValidationError("Can't handle RML samples")

        if not self._sample_helper.is_externally_sequenced(sample["internal_id"]):
            self._collect_missing_attributes(
                sample,
                [
                    "prepared_at",
                    "prep_method",
                    "sequencing_method",
                    "sequenced_at",
                    "million_read_pairs",
                    "mapped_reads",
                    "duplicates",
                ],
            )

        if self._sample_helper.is_analysis_sample(case_id):
            self._collect_missing_attributes(sample, REQUIRED_ANALYSIS_SAMPLE_FIELDS)

    def _check_required_general_report_data(self, report_data: dict) -> None:
        """Check root-level attributes for a report data set"""
        self._collect_missing_attributes(report_data, REQUIRED_REPORT_FIELDS)



"""Fixtures for cli tests"""
from functools import partial

import pytest
from click.testing import CliRunner

from cg.cli import base
from cg.store import Store


@pytest.fixture(name="cli_runner")
def fixture_cli_runner():
    """Create a CliRunner"""
    runner = CliRunner()
    return runner


@pytest.fixture
def invoke_cli(cli_runner):
    """Create a CLiRunner with our CLI preloaded"""
    return partial(cli_runner.invoke, base)


@pytest.fixture
def base_context(base_store: Store) -> dict:
    """context to use in cli"""
    return {"trailblazer_api": None, "status_db": base_store}



"""Handler to find basic data objects"""
import datetime as dt
from typing import List

from sqlalchemy import desc

from cg.store.api.base import BaseHandler
from cg.store import models


class FindBasicDataHandler(BaseHandler):
    """Contains methods to find basic data model instances"""

    def application(self, tag: str) -> models.Application:
        """Fetch an application from the store."""
        return self.Application.query.filter_by(tag=tag).first()

    def applications(self, *, category=None, archived=None):
        """Fetch all applications."""
        records = self.Application.query
        if category:
            records = records.filter_by(prep_category=category)
        if archived is not None:
            records = records.filter_by(is_archived=archived)
        return records.order_by(self.Application.prep_category, self.Application.tag)

    def application_version(
        self, application: models.Application, version: int
    ) -> models.ApplicationVersion:
        """Fetch an application version."""
        query = self.ApplicationVersion.query.filter_by(application=application, version=version)
        return query.first()

    def bed(self, name):
        """Find a bed by name."""
        return self.Bed.query.filter_by(name=name).first()

    def bed_version(self, shortname):
        """Find a bed version by shortname."""
        return self.BedVersion.query.filter_by(shortname=shortname).first()

    def beds(self, hide_archived: bool = False):
        """Returns all beds."""
        bed_q = self.Bed.query
        if hide_archived:
            bed_q = bed_q.filter(self.Bed.is_archived.is_(False))

        return bed_q.order_by(models.Bed.name)

    def latest_bed_version(self, name: str) -> models.BedVersion:
        """Fetch the latest bed version for a bed name."""
        bed_obj = self.Bed.query.filter_by(name=name).first()
        return bed_obj.versions[-1] if bed_obj and bed_obj.versions else None

    def customer(self, internal_id: str) -> models.Customer:
        """Fetch a customer by internal id from the store."""
        return self.Customer.query.filter_by(internal_id=internal_id).first()

    def customers(self) -> List[models.Customer]:
        """Fetch all customers."""
        return self.Customer.query

    def customer_group(self, internal_id: str) -> models.CustomerGroup:
        """Fetch a customer group by internal id from the store."""
        return self.CustomerGroup.query.filter_by(internal_id=internal_id).first()

    def customer_by_id(self, id_: int) -> models.Customer:
        """Fetch a customer by id number from the store."""
        return self.Customer.query.filter_by(id=id_).first()

    def current_application_version(self, tag: str) -> models.ApplicationVersion:
        """Fetch the current application version for an application tag."""
        application_obj = self.Application.query.filter_by(tag=tag).first()
        if not application_obj:
            return None
        application_id = application_obj.id
        records = self.ApplicationVersion.query.filter_by(application_id=application_id)
        records = records.filter(self.ApplicationVersion.valid_from < dt.datetime.now())
        records = records.order_by(desc(self.ApplicationVersion.valid_from))

        return records.first()

    def latest_version(self, tag: str) -> models.ApplicationVersion:
        """Fetch the latest application version for an application tag."""
        application_obj = self.Application.query.filter_by(tag=tag).first()
        return (
            application_obj.versions[-1] if application_obj and application_obj.versions else None
        )

    def organism(self, internal_id: str) -> models.Organism:
        """Find an Organism by internal_id."""
        return self.Organism.query.filter_by(internal_id=internal_id).first()

    def organisms(self) -> List[models.Organism]:
        """Fetch all organisms"""
        return self.Organism.query.order_by(models.Organism.internal_id)

    def panel(self, abbrev):
        """Find a panel by abbreviation."""
        return self.Panel.query.filter_by(abbrev=abbrev).first()

    def panels(self):
        """Returns all panels."""
        return self.Panel.query.order_by(models.Panel.abbrev)

    def user(self, email: str) -> models.User:
        """Fetch a user from the store."""
        return self.User.query.filter_by(email=email).first()



"""Test FastqFileConcatenator"""
from cg.apps.balsamic.fastq import FastqFileConcatenator


def test_concatenate(tmpdir, simple_files, files_content):
    """Test method to test that files are concatenated properly"""

    # given we have some files to concatenate and somewhere to store the concatenated file
    assert len(tmpdir.listdir()) == len(simple_files)
    concatenated_filepath = tmpdir + "/concatenated.fastq.gz"

    # when calling the method to concatenate
    FastqFileConcatenator.concatenate(simple_files, concatenated_filepath)

    # then we get a new file that is the concatenation of the others
    assert len(tmpdir.listdir()) == len(simple_files) + 1

    with open(concatenated_filepath, "rt") as file:
        file_content = file.read()

    assert files_content == file_content



"""This file tests the analyses_to_clean part of the status api"""
from datetime import datetime

from cg.constants import Pipeline
from cg.store import Store


def test_analysis_included(analysis_store: Store, helpers):
    """Tests that analyses that are uploaded are returned"""

    # GIVEN an analysis that is uploaded
    timestamp = datetime.now()
    analysis = helpers.add_analysis(
        analysis_store, started_at=timestamp, uploaded_at=timestamp, cleaned_at=None
    )
    sample = helpers.add_sample(analysis_store, delivered_at=timestamp)
    analysis_store.relate_sample(family=analysis.family, sample=sample, status="unknown")

    # WHEN calling the analyses_to_clean
    analyses_to_clean = analysis_store.analyses_to_clean(before=analysis.started_at).all()

    # THEN this analyse should be returned
    assert analysis in analyses_to_clean


def test_analysis_excluded(analysis_store: Store, helpers):
    """Tests that analyses that are completed but lacks delivery report are returned"""

    # GIVEN an analysis that is not uploaded
    timestamp = datetime.now()
    analysis = helpers.add_analysis(
        analysis_store, started_at=timestamp, uploaded_at=None, cleaned_at=None
    )
    sample = helpers.add_sample(analysis_store, delivered_at=timestamp)
    analysis_store.relate_sample(family=analysis.family, sample=sample, status="unknown")

    # WHEN calling the analyses_to_clean
    analyses_to_clean = analysis_store.analyses_to_clean().all()

    # THEN this analyse should be returned
    assert analysis not in analyses_to_clean


def test_pipeline_included(analysis_store: Store, helpers):
    """Tests that analyses that are included depending on pipeline"""

    # GIVEN an analysis that is uploaded and pipeline is specified
    timestamp = datetime.now()
    pipeline = Pipeline.BALSAMIC
    analysis = helpers.add_analysis(
        analysis_store,
        pipeline=pipeline,
        started_at=timestamp,
        uploaded_at=timestamp,
        cleaned_at=None,
    )
    sample = helpers.add_sample(analysis_store, delivered_at=timestamp)
    analysis_store.relate_sample(family=analysis.family, sample=sample, status="unknown")

    # WHEN calling the analyses_to_clean specifying the used pipeline
    analyses_to_clean = analysis_store.analyses_to_clean(pipeline=pipeline, before=timestamp).all()

    # THEN this analyse should be returned
    assert analysis in analyses_to_clean


def test_pipeline_excluded(analysis_store: Store, helpers):
    """Tests that analyses are excluded depending on pipeline"""

    # GIVEN an analysis that is uploaded
    timestamp = datetime.now()
    used_pipeline = Pipeline.BALSAMIC
    wrong_pipeline = Pipeline.MIP_DNA
    analysis = helpers.add_analysis(
        analysis_store,
        pipeline=used_pipeline,
        started_at=timestamp,
        uploaded_at=timestamp,
        cleaned_at=None,
    )
    sample = helpers.add_sample(analysis_store, delivered_at=timestamp)
    analysis_store.relate_sample(family=analysis.family, sample=sample, status="unknown")

    # WHEN calling the analyses_to_clean specifying another pipeline
    analyses_to_clean = analysis_store.analyses_to_clean(pipeline=wrong_pipeline).all()

    # THEN this analyse should not be returned
    assert analysis not in analyses_to_clean


def test_non_cleaned_included(analysis_store: Store, helpers):
    """Tests that analyses that are included depending on cleaned_at"""

    # GIVEN an analysis that is uploaded but not cleaned
    timestamp = datetime.now()
    analysis = helpers.add_analysis(
        analysis_store, started_at=timestamp, uploaded_at=timestamp, cleaned_at=None
    )
    sample = helpers.add_sample(analysis_store, delivered_at=timestamp)
    analysis_store.relate_sample(family=analysis.family, sample=sample, status="unknown")

    # WHEN calling the analyses_to_clean
    analyses_to_clean = analysis_store.analyses_to_clean(before=timestamp).all()

    # THEN this analyse should be returned
    assert analysis in analyses_to_clean


def test_cleaned_excluded(analysis_store: Store, helpers):
    """Tests that analyses are excluded depending on cleaned_at"""

    # GIVEN an analysis that is cleaned
    timestamp = datetime.now()
    analysis = helpers.add_analysis(
        analysis_store, started_at=timestamp, uploaded_at=timestamp, cleaned_at=timestamp
    )
    sample = helpers.add_sample(analysis_store, delivered_at=timestamp)
    analysis_store.relate_sample(family=analysis.family, sample=sample, status="unknown")

    # WHEN calling the analyses_to_clean
    analyses_to_clean = analysis_store.analyses_to_clean().all()

    # THEN this analyse should not be returned
    assert analysis not in analyses_to_clean



import logging
import os.path
from typing import Optional

import requests
from cg.exc import TicketCreationError
from flask import Flask

LOG = logging.getLogger(__name__)


class OsTicket(object):

    """Interface to ticket system."""

    def __init__(self):
        self.headers = None
        self.url = None

    def init_app(self, app: Flask):
        """Initialize the API in Flask."""
        self.setup(api_key=app.config["OSTICKET_API_KEY"], domain=app.config["OSTICKET_DOMAIN"])

    def setup(self, api_key: str = None, domain: str = None):
        """Initialize the API."""
        self.headers = {"X-API-Key": api_key}
        self.url = os.path.join(domain, "api/tickets.json")

    def open_ticket(self, name: str, email: str, subject: str, message: str) -> Optional[int]:
        """Open a new ticket through the REST API."""
        data = dict(name=name, email=email, subject=subject, message=message)
        res = requests.post(self.url, json=data, headers=self.headers)
        if res.ok:
            try:
                return int(res.text)
            except ValueError:
                LOG.error("Could not convert res %s to int", res.text)
                return None
        else:
            LOG.error("res.text: %s, reason: %s", res.text, res.reason)
            raise TicketCreationError(res)



from typing import List
from pathlib import Path
import ruamel
from cg.apps.tb.models import TrailblazerAnalysis


def get_sample_data_from_config(config: dict) -> List[dict]:
    """Get sample data from config"""
    samples = []
    for sample_id, analysis_type in config["analysis_type"].items():
        samples.append({"id": sample_id, "type": analysis_type})
    return samples


def parse_config(data: dict) -> dict:
    """Parse MIP config file

    Args:
        data (dict): raw YAML input from MIP analysis config file

    Returns:
        dict: parsed data
    """
    return {
        "email": data.get("email"),
        "case": data.get("case_id") or data.get("family_id"),
        "samples": get_sample_data_from_config(config=data),
        "config_path": data.get("config_file_analysis"),
        "is_dryrun": "dry_run_all" in data,
        "log_path": data.get("log_file"),
        "out_dir": data.get("outdata_dir"),
        "priority": data.get("slurm_quality_of_service"),
        "sampleinfo_path": data.get("sample_info_file"),
    }


def parse_sampleinfo_light(data: dict) -> dict:
    """Parse MIP sample info file and retrieve mip_version, date, and status

    Args:
        data (dict): raw YAML input from MIP qc sample info file

    Returns:
        dict: {'version': str, 'date': str, 'is_finished': str}


    """

    outdata = {
        "date": data.get("analysis_date"),
        "version": data.get("mip_version"),
        "is_finished": data.get("analysisrunstatus") == "finished",
    }

    return outdata


def get_rank_model_version(sample_info: dict, rank_model_type: str, step: str) -> str:
    """Get rank model version"""
    for key in ("recipe", "program"):
        if key in sample_info:
            return sample_info[key][step][rank_model_type]["version"]


def get_genome_build(sample_info: dict) -> str:
    """Get genome build"""
    genome_build = sample_info["human_genome_build"]
    return f"{genome_build['source']}{genome_build['version']}"


def parse_sampleinfo(data: dict) -> dict:
    """Parse MIP sample info file.

    Args:
        data (dict): raw YAML input from MIP qc sample info file

    Returns:
        dict: parsed data
    """
    outdata = {
        "date": data.get("analysis_date"),
        "genome_build": get_genome_build(sample_info=data),
        "case": data.get("case_id") or data.get("family_id"),
        "is_finished": data.get("analysisrunstatus") == "finished",
        "rank_model_version": get_rank_model_version(
            sample_info=data, rank_model_type="rank_model", step="genmod"
        ),
        "samples": [{"id": sample_id} for sample_id in data["sample"].items()],
        "sv_rank_model_version": get_rank_model_version(
            sample_info=data, rank_model_type="sv_rank_model", step="sv_genmod"
        ),
        "version": data["mip_version"],
    }

    return outdata


def parse_sampleinfo_rna(data: dict) -> dict:
    """Parse MIP sample info file (RNA).

    Args:
        data (dict): raw YAML input from MIP qc sample info file (RNA)

    Returns:
        dict: parsed data
    """
    case = data["case"]
    qcmetrics_parse = data["recipe"]["qccollect_ar"]

    outdata = {
        "date": data["analysis_date"],
        "is_finished": data["analysisrunstatus"] == "finished",
        "case": case,
        "config_file_path": data["config_file_analysis"],
        "version": data["mip_version"],
        "pedigree_path": data["pedigree_minimal"],
        "bcftools_merge": data["recipe"]["bcftools_merge"]["path"],
        "multiqc_html": data["recipe"]["multiqc"][f"{case}_html"]["path"],
        "multiqc_json": data["recipe"]["multiqc"][f"{case}_json"]["path"],
        "qcmetrics_path": qcmetrics_parse if not isinstance(qcmetrics_parse, dict) else None,
        "vep_path": data["recipe"]["varianteffectpredictor"]["path"],
        "version_collect_ar_path": data["recipe"]["version_collect_ar"]["path"],
        "samples": [],
    }

    for sample_id, sample_data in data["sample"].items():

        sample = {
            "id": sample_id,
            "bam": sample_data["most_complete_bam"]["path"],
            "star_fusion": sample_data["recipe"]["star_fusion"]["path"],
            "bootstrap_vcf": get_multiple_paths(sample_data, "bootstrapann"),
            "gatk_asereadcounter": get_multiple_paths(sample_data, "gatk_asereadcounter"),
            "gatk_baserecalibration": get_multiple_paths(sample_data, "gatk_baserecalibration"),
            "gffcompare_ar": get_multiple_paths(sample_data, "gffcompare_ar"),
            "mark_duplicates": get_multiple_paths(sample_data, "markduplicates"),
            "salmon_quant": get_multiple_paths(sample_data, "salmon_quant"),
            "stringtie_ar": get_multiple_paths(sample_data, "stringtie_ar"),
        }

        outdata["samples"].append(sample)

    return outdata


def get_multiple_paths(sample_data: dict, path_key: str) -> list:
    """Get all paths to files of a given type. Use this method if the exact filename is not
    known beforehand.
    Args:
        sample_data (dict): YAML file containing qc sample info
        path_key (str)    : Type of file paths to retrieve

    Returns:
        list: paths to all files of given type
    """
    paths = [file_data["path"] for file_data in sample_data["recipe"][path_key].values()]

    return paths


def get_sampleinfo(analysis_obj: TrailblazerAnalysis) -> str:
    """Get the sample info path for an analysis."""
    raw_data = ruamel.yaml.safe_load(Path(analysis_obj.config_path).open())
    data = parse_config(raw_data)
    return data["sampleinfo_path"]



import datetime as dt

import pytest
from cg.exc import OrderError
from cg.meta.orders import OrdersAPI, OrderType
from cg.store import models


@pytest.mark.parametrize(
    "order_type",
    [
        OrderType.RML,
        OrderType.FASTQ,
        OrderType.MIP_DNA,
        OrderType.MIP_RNA,
        OrderType.EXTERNAL,
        OrderType.MICROSALT,
        OrderType.METAGENOME,
        OrderType.BALSAMIC,
    ],
)
def test_submit(
    base_store,
    orders_api: OrdersAPI,
    all_orders_to_submit,
    monkeypatch,
    order_type,
    user_name: str,
    user_mail: str,
    ticket_number: int,
):
    order_data = all_orders_to_submit[order_type]
    lims_project_data = {"id": "ADM1234", "date": dt.datetime.now()}
    lims_map = {
        sample["name"]: f"ELH123A{index}" for index, sample in enumerate(order_data.samples)
    }
    monkeypatch.setattr(orders_api, "process_lims", lambda *args: (lims_project_data, lims_map))

    # GIVEN an order and an empty store
    assert base_store.samples().first() is None

    # WHEN submitting the order

    result = orders_api.submit(
        project=order_type, order_in=order_data, user_name=user_name, user_mail=user_mail
    )

    # THEN the result should contain the ticket number for the order
    for record in result["records"]:
        if isinstance(record, (models.Pool, models.Sample)):
            assert record.ticket_number == ticket_number
        elif isinstance(record, models.Family):
            for link_obj in record.links:
                assert link_obj.sample.ticket_number == ticket_number


@pytest.mark.parametrize(
    "order_type",
    [OrderType.MIP_DNA, OrderType.MIP_RNA, OrderType.EXTERNAL, OrderType.BALSAMIC],
)
def test_submit_illegal_sample_customer(
    sample_store,
    orders_api,
    all_orders_to_submit,
    monkeypatch,
    order_type,
    ticket_number: int,
    user_name: str,
    user_mail: str,
):

    order_data = all_orders_to_submit[order_type]
    lims_project_data = {"id": "ADM1234", "date": dt.datetime.now()}
    lims_map = {
        sample["name"]: f"ELH123A{index}" for index, sample in enumerate(order_data.samples)
    }
    monkeypatch.setattr(orders_api, "process_lims", lambda *args: (lims_project_data, lims_map))

    # GIVEN we have an order with a customer that is not in the same customer group as customer
    # that the samples originate from
    customer_group = sample_store.add_customer_group("customer999only", "customer 999 only group")
    sample_store.add_commit(customer_group)
    new_customer = sample_store.add_customer(
        "customer999",
        "customer 999",
        scout_access=True,
        invoice_address="dummy street",
        customer_group=customer_group,
        invoice_reference="dummy nr",
    )
    sample_store.add_commit(new_customer)
    existing_sample = sample_store.samples().first()
    existing_sample.customer = new_customer
    sample_store.add_commit(existing_sample)

    for sample in order_data.samples:
        sample["internal_id"] = existing_sample.internal_id

    # WHEN calling submit
    # THEN an OrderError should be raised on illegal customer
    with pytest.raises(OrderError):
        orders_api.submit(
            project=order_type, order_in=order_data, user_name=user_name, user_mail=user_mail
        )


@pytest.mark.parametrize(
    "order_type",
    [OrderType.MIP_DNA, OrderType.MIP_RNA, OrderType.EXTERNAL, OrderType.BALSAMIC],
)
def test_submit_scout_legal_sample_customer(
    sample_store,
    orders_api,
    all_orders_to_submit,
    monkeypatch,
    order_type,
    user_name: str,
    user_mail: str,
    ticket_number: int,
):

    order_data = all_orders_to_submit[order_type]
    lims_project_data = {"id": "ADM1234", "date": dt.datetime.now()}
    lims_map = {
        sample["name"]: f"ELH123A{index}" for index, sample in enumerate(order_data.samples)
    }
    monkeypatch.setattr(orders_api, "process_lims", lambda *args: (lims_project_data, lims_map))
    # GIVEN we have an order with a customer that is in the same customer group as customer
    # that the samples originate from
    customer_group = sample_store.add_customer_group("customer999only", "customer 999 only group")
    sample_store.add_commit(customer_group)
    sample_customer = sample_store.add_customer(
        "customer1",
        "customer 1",
        scout_access=True,
        invoice_address="dummy street 1",
        customer_group=customer_group,
        invoice_reference="dummy nr",
    )
    order_customer = sample_store.add_customer(
        "customer2",
        "customer 2",
        scout_access=True,
        invoice_address="dummy street 2",
        customer_group=customer_group,
        invoice_reference="dummy nr",
    )
    sample_store.add_commit(sample_customer)
    sample_store.add_commit(order_customer)
    existing_sample = sample_store.samples().first()
    existing_sample.customer = sample_customer
    sample_store.commit()
    order_data.customer = order_customer.internal_id
    first = True

    for sample in order_data.samples:
        if first:
            sample["internal_id"] = existing_sample.internal_id
            first = False

    # WHEN calling submit
    # THEN an OrderError should not be raised on illegal customer
    orders_api.submit(
        project=order_type, order_in=order_data, user_name=user_name, user_mail=user_mail
    )


@pytest.mark.parametrize(
    "order_type",
    [OrderType.RML, OrderType.FASTQ, OrderType.MICROSALT, OrderType.METAGENOME],
)
def test_submit_non_scout_legal_sample_customer(
    sample_store,
    orders_api,
    all_orders_to_submit,
    monkeypatch,
    order_type,
    user_name: str,
    user_mail: str,
    ticket_number: int,
):

    order_data = all_orders_to_submit[order_type]
    lims_project_data = {"id": "ADM1234", "date": dt.datetime.now()}
    lims_map = {
        sample["name"]: f"ELH123A{index}" for index, sample in enumerate(order_data.samples)
    }
    monkeypatch.setattr(orders_api, "process_lims", lambda *args: (lims_project_data, lims_map))

    # GIVEN we have an order with a customer that is in the same customer group as customer
    # that the samples originate from but on order types where this is dis-allowed
    customer_group = sample_store.add_customer_group("customer999only", "customer 999 only group")
    sample_store.add_commit(customer_group)
    sample_customer = sample_store.add_customer(
        "customer1",
        "customer 1",
        scout_access=True,
        invoice_address="dummy street 1",
        customer_group=customer_group,
        invoice_reference="dummy nr",
    )
    order_customer = sample_store.add_customer(
        "customer2",
        "customer 2",
        scout_access=True,
        invoice_address="dummy street 2",
        customer_group=customer_group,
        invoice_reference="dummy nr",
    )
    sample_store.add_commit(sample_customer)
    sample_store.add_commit(order_customer)
    existing_sample = sample_store.samples().first()
    existing_sample.customer = sample_customer
    sample_store.commit()
    order_data.customer = order_customer.internal_id
    first = True

    for sample in order_data.samples:
        if first:
            sample["internal_id"] = existing_sample.internal_id
            first = False

    # WHEN calling submit
    # THEN an OrderError should be raised on illegal customer
    with pytest.raises(OrderError):
        orders_api.submit(
            project=order_type, order_in=order_data, user_name=user_name, user_mail=user_mail
        )



from typing import Iterable, List

import pytest

from cg.store import Store, models
from cg.store.api.import_func import (
    parse_application_versions,
    parse_applications,
    versions_are_same,
)
from cg.store.api.models import ApplicationSchema, ApplicationVersionSchema
from tests.store_helpers import StoreHelpers


class StoreCheckers:
    @staticmethod
    def get_versions_from_store(
        store: Store, application_tag: str
    ) -> List[models.ApplicationVersion]:
        """Gets all versions for the specified application"""

        return store.application(application_tag).versions

    @staticmethod
    def get_application_from_store(store: Store, application_tag: str) -> models.Application:
        """Gets the specified application"""

        return store.application(application_tag)

    @staticmethod
    def version_exists_in_store(store: Store, application: ApplicationVersionSchema):
        """Check if the given raw version exists in the store"""
        db_versions: List[models.Application] = StoreCheckers.get_versions_from_store(
            store=store, application_tag=application.app_tag
        )

        for db_version in db_versions:
            if versions_are_same(version_obj=db_version, application_version=application):
                return True

        return False

    @staticmethod
    def all_versions_exist_in_store(store: Store, excel_path: str):
        """Check if all versions in the excel exists in the store"""
        applications: Iterable[ApplicationVersionSchema] = parse_application_versions(
            excel_path=excel_path
        )

        for application in applications:
            if not StoreCheckers.version_exists_in_store(store=store, application=application):
                return False

        return True

    @staticmethod
    def all_applications_exists(store: Store, applications_file: str):
        """Check if all applications in the excel exists in the store"""
        applications: Iterable[ApplicationSchema] = parse_applications(excel_path=applications_file)

        for application in applications:
            if not StoreCheckers.exists_application_in_store(
                store=store, application_tag=application.tag
            ):
                return False

        return True

    @staticmethod
    def exists_application_in_store(store: Store, application_tag: str):
        """Check if the given raw application exists in the store"""
        db_application = StoreCheckers.get_application_from_store(
            store=store, application_tag=application_tag
        )

        return db_application is not None


@pytest.fixture(name="store_checkers")
def fixture_store_checkers() -> StoreCheckers:
    return StoreCheckers()


@pytest.fixture(name="applications_store")
def fixture_applications_store(
    store: Store, application_versions_file: str, helpers: StoreHelpers
) -> Store:
    """Return a store populated with applications from excel file"""
    versions: Iterable[ApplicationVersionSchema] = parse_application_versions(
        excel_path=application_versions_file
    )

    for version in versions:
        helpers.ensure_application(store=store, tag=version.app_tag)

    return store


@pytest.fixture(name="microbial_store")
def fixture_microbial_store(store: Store, helpers: StoreHelpers) -> Store:
    """Populate a store with microbial application tags"""
    microbial_active_apptags = ["MWRNXTR003", "MWGNXTR003", "MWMNXTR003", "MWLNXTR003"]
    microbial_inactive_apptags = ["MWXNXTR003", "VWGNXTR001", "VWLNXTR001"]

    for app_tag in microbial_active_apptags:
        helpers.ensure_application(
            store=store, tag=app_tag, application_type="mic", is_archived=False
        )

    for app_tag in microbial_inactive_apptags:
        helpers.ensure_application(
            store=store, tag=app_tag, application_type="mic", is_archived=True
        )

    return store


@pytest.fixture(name="microbial_store_dummy_tag")
def fixture_microbial_store_dummy_tag(microbial_store: Store, helpers: StoreHelpers) -> Store:
    """Populate a microbial store with a extra dummy app tag"""
    helpers.ensure_application(
        store=microbial_store, tag="dummy_tag", application_type="mic", is_archived=False
    )
    return microbial_store



"""Code that handles CLI commands to upload"""
import datetime as dt
import logging
import sys
import traceback

import click
from cg.apps.coverage import ChanjoAPI
from cg.apps.gt import GenotypeAPI
from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.apps.lims import LimsAPI
from cg.apps.madeline.api import MadelineAPI
from cg.apps.scout.scoutapi import ScoutAPI
from cg.apps.tb import TrailblazerAPI
from cg.constants import Pipeline
from cg.exc import AnalysisUploadError
from cg.meta.report.api import ReportAPI
from cg.meta.upload.scout.scoutapi import UploadScoutAPI
from cg.meta.workflow.mip import MipAnalysisAPI
from cg.store import Store
from cg.utils.click.EnumChoice import EnumChoice

from .coverage import coverage
from .delivery_report import delivery_report, delivery_report_to_scout, delivery_reports
from .genotype import genotypes
from .mutacc import process_solved, processed_solved
from .observations import observations
from .scout import create_scout_load_config, scout, upload_case_to_scout
from .utils import suggest_cases_to_upload
from .validate import validate

LOG = logging.getLogger(__name__)


@click.group(invoke_without_command=True)
@click.option("-f", "--family", "family_id", help="Upload to all apps")
@click.option(
    "-r",
    "--restart",
    "force_restart",
    is_flag=True,
    help="Force upload of analysis " "marked as started",
)
@click.pass_context
def upload(context, family_id, force_restart):
    """Upload results from analyses."""

    click.echo(click.style("----------------- UPLOAD ----------------------"))

    context.obj["status_db"] = Store(context.obj["database"])

    if family_id:
        case_obj = context.obj["status_db"].family(family_id)
        if not case_obj:
            message = f"family not found: {family_id}"
            click.echo(click.style(message, fg="red"))
            context.abort()

        if not case_obj.analyses:
            message = f"no analysis exists for family: {family_id}"
            click.echo(click.style(message, fg="red"))
            context.abort()

        analysis_obj = case_obj.analyses[0]

        if analysis_obj.uploaded_at is not None:
            message = f"analysis already uploaded: {analysis_obj.uploaded_at.date()}"
            click.echo(click.style(message, fg="red"))
            context.abort()

        if not force_restart and analysis_obj.upload_started_at is not None:
            if dt.datetime.now() - analysis_obj.upload_started_at > dt.timedelta(hours=24):
                raise AnalysisUploadError(
                    f"The upload started at {analysis_obj.upload_started_at} "
                    f"something went wrong, restart it with the --restart flag"
                )

            message = f"analysis upload already started: {analysis_obj.upload_started_at.date()}"
            click.echo(click.style(message, fg="yellow"))
            return

    context.obj["housekeeper_api"] = HousekeeperAPI(context.obj)
    context.obj["madeline_api"] = MadelineAPI(context.obj)
    context.obj["genotype_api"] = GenotypeAPI(context.obj)
    context.obj["lims_api"] = LimsAPI(context.obj)
    context.obj["trailblazer_api"] = TrailblazerAPI(context.obj)
    context.obj["chanjo_api"] = ChanjoAPI(context.obj)
    context.obj["scout_api"] = ScoutAPI(context.obj)

    context.obj["analysis_api"] = MipAnalysisAPI(
        db=context.obj["status_db"],
        hk_api=context.obj["housekeeper_api"],
        tb_api=context.obj["trailblazer_api"],
        scout_api=context.obj["scout_api"],
        lims_api=context.obj["lims_api"],
        script=context.obj["mip-rd-dna"]["script"],
        pipeline=context.obj["mip-rd-dna"]["pipeline"],
        conda_env=context.obj["mip-rd-dna"]["conda_env"],
        root=context.obj["mip-rd-dna"]["root"],
    )
    context.obj["report_api"] = ReportAPI(
        store=context.obj["status_db"],
        lims_api=context.obj["lims_api"],
        chanjo_api=context.obj["chanjo_api"],
        analysis_api=context.obj["analysis_api"],
        scout_api=context.obj["scout_api"],
    )

    context.obj["scout_upload_api"] = UploadScoutAPI(
        hk_api=context.obj["housekeeper_api"],
        scout_api=context.obj["scout_api"],
        madeline_api=context.obj["madeline_api"],
        analysis_api=context.obj["analysis_api"],
        lims_api=context.obj["lims_api"],
    )

    if context.invoked_subcommand is not None:
        return

    if not family_id:
        suggest_cases_to_upload(context)
        context.abort()

    case_obj = context.obj["status_db"].family(family_id)
    analysis_obj = case_obj.analyses[0]
    if analysis_obj.uploaded_at is not None:
        message = f"analysis already uploaded: {analysis_obj.uploaded_at.date()}"
        click.echo(click.style(message, fg="yellow"))
    else:
        analysis_obj.upload_started_at = dt.datetime.now()
        context.obj["status_db"].commit()
        context.invoke(coverage, re_upload=True, family_id=family_id)
        context.invoke(validate, family_id=family_id)
        context.invoke(genotypes, re_upload=False, family_id=family_id)
        context.invoke(observations, case_id=family_id)
        context.invoke(scout, case_id=family_id)
        analysis_obj.uploaded_at = dt.datetime.now()
        context.obj["status_db"].commit()
        click.echo(click.style(f"{family_id}: analysis uploaded!", fg="green"))


@upload.command()
@click.option("--pipeline", type=EnumChoice(Pipeline), help="Limit to specific pipeline")
@click.pass_context
def auto(context: click.Context, pipeline: Pipeline = None):
    """Upload all completed analyses."""

    click.echo(click.style("----------------- AUTO ------------------------"))

    exit_code = 0
    for analysis_obj in context.obj["status_db"].analyses_to_upload(pipeline=pipeline):

        if analysis_obj.family.analyses[0].uploaded_at is not None:
            LOG.warning(
                "Newer analysis already uploaded for %s, skipping",
                analysis_obj.family.internal_id,
            )
            continue
        internal_id = analysis_obj.family.internal_id

        LOG.info("Uploading family: %s", internal_id)
        try:
            context.invoke(upload, family_id=internal_id)
        except Exception:
            LOG.error("Uploading family failed: %s", internal_id)
            LOG.error(traceback.format_exc())
            exit_code = 1

    sys.exit(exit_code)


upload.add_command(process_solved)
upload.add_command(processed_solved)
upload.add_command(validate)
upload.add_command(scout)
upload.add_command(upload_case_to_scout)
upload.add_command(create_scout_load_config)
upload.add_command(observations)
upload.add_command(genotypes)
upload.add_command(coverage)
upload.add_command(delivery_report)
upload.add_command(delivery_reports)
upload.add_command(delivery_report_to_scout)



import logging
from typing import Callable

import petname
from cg.store import models

LOG = logging.getLogger(__name__)


def case_exists(case_obj, case_id: str):
    """
    Check that case exist in Store
    """
    if case_obj is None:
        LOG.error("%s: case not found", case_id)
        return None
    return True


def get_links(case: models.Family) -> models.Sample:
    """Return all samples from a family object"""
    yield from case.links


def get_samples(analysis_obj: models.Analysis) -> models.Sample:
    """Search a analysis object and return the samples"""
    return get_links(analysis_obj.family)


def get_unique_id(availability_func: Callable) -> str:
    """Generate a unique sample id.

    The availability function should return a matching result given an id if it
    already exists.
    """
    while True:
        random_id = petname.Generate(3, separator="")
        if availability_func(random_id) is None:
            return random_id
        else:
            LOG.debug(f"{random_id} already used - trying another id")


def reset_case_action(case_obj):
    """ Resets action on case """

    case_obj.action = None



import logging
import click

from cg.apps.hermes.hermes_api import HermesApi
from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.apps.lims import LimsAPI
from cg.apps.tb import TrailblazerAPI
from cg.store import Store
from cg.constants import EXIT_SUCCESS, EXIT_FAIL, Pipeline
from cg.apps.environ import environ_email
from cg.meta.workflow.fluffy import FluffyAnalysisAPI

OPTION_DRY = click.option(
    "-d", "--dry-run", "dry_run", help="Print command to console without executing", is_flag=True
)
ARGUMENT_CASE_ID = click.argument("case_id", required=True)

LOG = logging.getLogger(__name__)


@click.group(invoke_without_command=True)
@click.pass_context
def fluffy(context: click.Context):
    """
    Fluffy workflow
    """
    if context.invoked_subcommand is None:
        LOG.info(context.get_help())
        return None
    config = context.obj
    context.obj["fluffy_analysis_api"] = FluffyAnalysisAPI(
        housekeeper_api=HousekeeperAPI(config),
        trailblazer_api=TrailblazerAPI(config),
        hermes_api=HermesApi(config),
        lims_api=LimsAPI(config),
        status_db=Store(config["database"]),
        config=config["fluffy"],
    )


@fluffy.command()
@ARGUMENT_CASE_ID
@OPTION_DRY
@click.pass_context
def link(context: click.Context, case_id: str, dry_run: bool):
    """
    Link fastq files from Housekeeper to analysis folder
    """
    fluffy_analysis_api: FluffyAnalysisAPI = context.obj["fluffy_analysis_api"]
    fluffy_analysis_api.verify_case_id_in_database(case_id=case_id)
    fluffy_analysis_api.link_fastq_files(case_id=case_id, dry_run=dry_run)


@fluffy.command("create-samplesheet")
@ARGUMENT_CASE_ID
@OPTION_DRY
@click.pass_context
def create_samplesheet(context: click.Context, case_id: str, dry_run: bool):
    """
    Write modified samplesheet file to case folder
    """
    fluffy_analysis_api: FluffyAnalysisAPI = context.obj["fluffy_analysis_api"]
    fluffy_analysis_api.verify_case_id_in_database(case_id=case_id)
    fluffy_analysis_api.make_samplesheet(case_id=case_id, dry_run=dry_run)


@fluffy.command()
@ARGUMENT_CASE_ID
@OPTION_DRY
@click.pass_context
def run(context: click.Context, case_id: str, dry_run: bool):
    """
    Run Fluffy analysis
    """
    fluffy_analysis_api: FluffyAnalysisAPI = context.obj["fluffy_analysis_api"]
    fluffy_analysis_api.verify_case_id_in_database(case_id=case_id)
    fluffy_analysis_api.run_fluffy(case_id=case_id, dry_run=dry_run)
    if dry_run:
        return
    # Submit analysis for tracking in Trailblazer
    try:
        fluffy_analysis_api.trailblazer_api.add_pending_analysis(
            case_id=case_id,
            email=environ_email(),
            type="tgs",
            out_dir=fluffy_analysis_api.get_output_path(case_id).as_posix(),
            config_path=fluffy_analysis_api.get_slurm_job_ids_path(case_id).as_posix(),
            priority=fluffy_analysis_api.get_priority(case_id),
            data_analysis="FLUFFY",
        )
        LOG.info("Submitted case %s to Trailblazer!", case_id)
    except Exception as e:
        LOG.warning("Unable to submit job file to Trailblazer, raised error: %s", e)

    fluffy_analysis_api.set_statusdb_action(case_id=case_id, action="running")


@fluffy.command()
@ARGUMENT_CASE_ID
@OPTION_DRY
@click.pass_context
def start(context: click.Context, case_id: str, dry_run: bool):
    """
    Starts full Fluffy analysis workflow
    """
    LOG.info("Starting full Fluffy workflow for %s", case_id)
    if dry_run:
        LOG.info("Dry run: the executed commands will not produce output!")
    context.invoke(link, case_id=case_id, dry_run=dry_run)
    context.invoke(create_samplesheet, case_id=case_id, dry_run=dry_run)
    context.invoke(run, case_id=case_id, dry_run=dry_run)


@fluffy.command("start-available")
@OPTION_DRY
@click.pass_context
def start_available(context: click.Context, dry_run: bool):
    """
    Start full Fluffy workflow for all cases/batches ready to be analyzed
    """
    exit_code = EXIT_SUCCESS
    fluffy_analysis_api: FluffyAnalysisAPI = context.obj["fluffy_analysis_api"]
    for case_obj in fluffy_analysis_api.status_db.cases_to_analyze(pipeline=Pipeline.FLUFFY):
        try:
            context.invoke(start, case_id=case_obj.internal_id, dry_run=dry_run)
        except Exception as exception_object:
            LOG.error(f"Exception occurred - {exception_object.__class__.__name__}")
            exit_code = EXIT_FAIL
    if exit_code:
        raise click.Abort()


@fluffy.command()
@ARGUMENT_CASE_ID
@OPTION_DRY
@click.pass_context
def store(context: click.Context, case_id: str, dry_run: bool):
    """
    Store finished analysis files in Housekeeper
    """
    fluffy_analysis_api: FluffyAnalysisAPI = context.obj["fluffy_analysis_api"]
    fluffy_analysis_api.verify_case_id_in_database(case_id=case_id)
    if dry_run:
        LOG.info("Dry run: Would have stored deliverables for %s", case_id)
        return
    try:
        fluffy_analysis_api.upload_bundle_housekeeper(case_id=case_id)
        fluffy_analysis_api.upload_bundle_statusdb(case_id=case_id)
        fluffy_analysis_api.set_statusdb_action(case_id=case_id, action=None)
    except Exception as error:
        fluffy_analysis_api.housekeeper_api.rollback()
        fluffy_analysis_api.status_db.rollback()
        LOG.error("Error storing deliverables for case %s - %s", case_id, error.__class__.__name__)
        raise


@fluffy.command("store-available")
@OPTION_DRY
@click.pass_context
def store_available(context: click.Context, dry_run: bool) -> None:
    """
    Store bundles for all finished analyses in Housekeeper
    """
    fluffy_analysis_api: FluffyAnalysisAPI = context.obj["fluffy_analysis_api"]
    exit_code: int = EXIT_SUCCESS
    for case_obj in fluffy_analysis_api.get_cases_to_store():
        LOG.info("Storing deliverables for %s", case_obj.internal_id)
        try:
            context.invoke(store, case_id=case_obj.internal_id, dry_run=dry_run)
        except Exception:
            exit_code = EXIT_FAIL
    if exit_code:
        raise click.Abort



""" Test the CLI for run mip-rna """
import logging

from cg.cli.workflow.mip_rna.base import run


def test_cg_dry_run(cli_runner, caplog, case_id, email_adress, rna_mip_context):
    """Test print the MIP command to console"""

    caplog.set_level(logging.INFO)
    # GIVEN a cli function

    # WHEN we run a case in dry run mode
    result = cli_runner.invoke(
        run, ["--dry-run", "--email", email_adress, case_id], obj=rna_mip_context
    )

    # THEN the command should be printed
    assert result.exit_code == 0
    assert f"analyse rd_rna {case_id} --config config.yaml --email {email_adress}" in caplog.text



""" API to manage Microsalt Analyses
    Organism - Fallback based on reference, ‘Other species’ and ‘Comment’. Default to “Unset”.
    Priority = Default to empty string. Weird response. Typically “standard” or “research”.
    Reference = Defaults to “None”
    Method: Outputted as “1273:23”. Defaults to “Not in LIMS”
    Date: Returns latest == most recent date. Outputted as DT object “YYYY MM DD”. Defaults to
    datetime.min"""
import gzip
import logging
import os
import re
from datetime import datetime
from pathlib import Path
from subprocess import CalledProcessError
from typing import Any, Dict, List, Optional

import click

from cg.apps.environ import environ_email
from cg.apps.hermes.hermes_api import HermesApi
from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.apps.lims import LimsAPI
from cg.apps.tb import TrailblazerAPI
from cg.constants import CASE_ACTIONS, Pipeline
from cg.exc import BundleAlreadyAddedError, CgDataError
from cg.store import Store, models
from cg.utils import Process

LOG = logging.getLogger(__name__)


class MicrosaltAnalysisAPI:
    """API to manage Microsalt Analyses"""

    def __init__(
        self,
        db: Store,
        hk_api: HousekeeperAPI,
        lims_api: LimsAPI,
        hermes_api: HermesApi,
        trailblazer_api: TrailblazerAPI,
        config: Optional[dict] = {},
    ):
        self.db = db
        self.hk = hk_api
        self.lims = lims_api
        self.hermes_api = hermes_api
        self.trailblazer_api = trailblazer_api
        self.root_dir = config["root"]
        self.queries_path = config["queries_path"]
        self.process = Process(binary=config["binary_path"], environment=config["conda_env"])

    def check_flowcells_on_disk(self, case_id: str, sample_id: Optional[str] = None) -> bool:
        """Check if flowcells are on disk for sample before starting the analysis.
        Flowcells not on disk will be requested
        """

        flowcells = self.get_flowcells(case_id=case_id, sample_id=sample_id)
        statuses = []
        for flowcell_obj in flowcells:
            LOG.debug(f"{flowcell_obj.name}: checking if flowcell is on disk")
            statuses.append(flowcell_obj.status)
            if flowcell_obj.status == "removed":
                LOG.info(f"{flowcell_obj.name}: flowcell not on disk, requesting")
                flowcell_obj.status = "requested"
            elif flowcell_obj.status != "ondisk":
                LOG.warning(f"{flowcell_obj.name}: {flowcell_obj.status}")
        self.db.commit()
        return all(status == "ondisk" for status in statuses)

    def get_flowcells(self, case_id: str, sample_id: Optional[str] = None) -> List[models.Flowcell]:
        """Get all flowcells for all samples in a ticket"""
        flowcells = set()
        for sample in self.get_samples(case_id=case_id, sample_id=sample_id):
            for flowcell in sample.flowcells:
                flowcells.add(flowcell)

        return list(flowcells)

    def get_case_fastq_path(self, case_id: str) -> Path:
        return Path(self.root_dir, "fastq", case_id)

    def get_config_path(self, filename: str) -> Path:
        return Path(self.queries_path, filename).with_suffix(".json")

    def get_trailblazer_config_path(self, case_id: str) -> Path:
        case_obj: models.Family = self.db.family(case_id)
        order_id = case_obj.name
        return Path(
            self.root_dir, "results", "reports", "trailblazer", f"{order_id}_slurm_ids.yaml"
        )

    def get_deliverables_file_path(self, case_id: str) -> Path:
        """Returns a path where the microSALT deliverables file for the order_id should be
        located"""
        case_obj: models.Family = self.db.family(case_id)
        order_id = case_obj.name
        deliverables_file_path = Path(
            self.root_dir,
            "results",
            "reports",
            "deliverables",
            f"{order_id}_deliverables.yaml",
        )
        if deliverables_file_path.exists():
            LOG.info("Found deliverables file %s", deliverables_file_path)
        return deliverables_file_path

    @staticmethod
    def generate_fastq_name(
        lane: str, flowcell: str, sample: str, read: str, more: dict = None
    ) -> str:
        """Name a FASTQ file following usalt conventions. Naming must be
        xxx_R_1.fastq.gz and xxx_R_2.fastq.gz"""

        undetermined = more.get("undetermined", None)
        # ACC1234A1_FCAB1ABC2_L1_1.fastq.gz sample_flowcell_lane_read.fastq.gz
        flowcell = f"{flowcell}-undetermined" if undetermined else flowcell
        return f"{sample}_{flowcell}_L{lane}_{read}.fastq.gz"

    @staticmethod
    def fastq_header(line: str) -> dict:
        """handle illumina's two different header formats
        @see https://en.wikipedia.org/wiki/FASTQ_format

        @HWUSI-EAS100R:6:73:941:1973#0/1

            HWUSI-EAS100R   the unique instrument name
            6   flowcell lane
            73  tile number within the flowcell lane
            941     'x'-coordinate of the cluster within the tile
            1973    'y'-coordinate of the cluster within the tile
            #0  index number for a multiplexed sample (0 for no indexing)
            /1  the member of a pair, /1 or /2 (paired-end or mate-pair reads only)

        Versions of the Illumina pipeline since 1.4 appear to use #NNNNNN
        instead of #0 for the multiplex ID, where NNNNNN is the sequence of the
        multiplex tag.

        With Casava 1.8 the format of the '@' line has changed:

        @EAS139:136:FC706VJ:2:2104:15343:197393 1:Y:18:ATCACG

            EAS139  the unique instrument name
            136     the run id
            FC706VJ     the flowcell id
            2   flowcell lane
            2104    tile number within the flowcell lane
            15343   'x'-coordinate of the cluster within the tile
            197393  'y'-coordinate of the cluster within the tile
            1   the member of a pair, 1 or 2 (paired-end or mate-pair reads only)
            Y   Y if the read is filtered, N otherwise
            18  0 when none of the control bits are on, otherwise it is an even number
            ATCACG  index sequence
        """

        rs = {"lane": None, "flowcell": None, "readnumber": None}

        parts = line.split(":")
        if len(parts) == 5:  # @HWUSI-EAS100R:6:73:941:1973#0/1
            rs["lane"] = parts[1]
            rs["flowcell"] = "XXXXXX"
            rs["readnumber"] = parts[-1].split("/")[-1]
        if len(parts) == 10:  # @EAS139:136:FC706VJ:2:2104:15343:197393 1:Y:18:ATCACG
            rs["lane"] = parts[3]
            rs["flowcell"] = parts[2]
            rs["readnumber"] = parts[6].split(" ")[-1]
        if len(parts) == 7:  # @ST-E00201:173:HCLCGALXX:1:2106:22516:34834/1
            rs["lane"] = parts[3]
            rs["flowcell"] = parts[2]
            rs["readnumber"] = parts[-1].split("/")[-1]

        return rs

    def link_samples(self, case_id: str, sample_id: Optional[str] = None) -> None:

        case_dir: Path = self.get_case_fastq_path(case_id=case_id)
        case_dir.mkdir(parents=True, exist_ok=True)

        samples = self.get_samples(case_id=case_id, sample_id=sample_id)

        for sample in samples:
            LOG.info("%s: link FASTQ files", sample.internal_id)
            self.link_fastq(
                case_dir,
                sample_id=sample.internal_id,
            )

    def link_fastq(self, case_dir: Path, sample_id: str) -> None:
        """Link FASTQ files for a sample."""

        fastq_dir = Path(case_dir, sample_id)
        fastq_dir.mkdir(exist_ok=True, parents=True)

        file_objs = self.hk.files(bundle=sample_id, tags=["fastq"])
        files = []

        for file_obj in file_objs:
            # figure out flowcell name from header
            with gzip.open(file_obj.full_path) as handle:
                header_line = handle.readline().decode()
                header_info = self.fastq_header(header_line)

            data = {
                "path": file_obj.full_path,
                "lane": int(header_info["lane"]),
                "flowcell": header_info["flowcell"],
                "read": int(header_info["readnumber"]),
                "undetermined": ("_Undetermined_" in file_obj.path),
            }
            # look for tile identifier (HiSeq X runs)
            matches = re.findall(r"-l[1-9]t([1-9]{2})_", file_obj.path)
            if len(matches) > 0:
                data["flowcell"] = f"{data['flowcell']}-{matches[0]}"
            files.append(data)

        for fastq_data in files:
            original_fastq_path = Path(fastq_data["path"])
            linked_fastq_name = self.generate_fastq_name(
                lane=fastq_data["lane"],
                flowcell=fastq_data["flowcell"],
                sample=sample_id,
                read=fastq_data["read"],
                more={"undetermined": fastq_data["undetermined"]},
            )
            linked_fastq_path = Path(fastq_dir, linked_fastq_name)

            if not linked_fastq_path.exists():
                LOG.info("Linking: %s -> %s", original_fastq_path, linked_fastq_path)
                linked_fastq_path.symlink_to(original_fastq_path)
            else:
                LOG.debug("Destination path already exists: %s, skipping", linked_fastq_path)

    def get_samples(self, case_id: str, sample_id: Optional[str] = None) -> List[models.Sample]:
        """Returns a list of samples to configure
        If sample_id is specified, will return a list with only this sample_id.
        Otherwise, returns all samples in given case"""
        if sample_id:
            return [
                self.db.query(models.Sample).filter(models.Sample.internal_id == sample_id).first()
            ]

        case_obj: models.Family = self.db.family(case_id)
        return [link.sample for link in case_obj.links]

    def get_lims_comment(self, sample_id: str) -> str:
        """ Returns the comment associated with a sample stored in lims"""
        comment = self.lims.get_sample_comment(sample_id) or ""
        if re.match(r"\w{4}\d{2,3}", comment):
            return comment

        return ""

    def get_organism(self, sample_obj: models.Sample) -> str:
        """Organism
        - Fallback based on reference, ‘Other species’ and ‘Comment’.
        Default to "Unset"."""

        if not sample_obj.organism:
            raise CgDataError(f"Organism missing on Sample")

        organism = sample_obj.organism.internal_id.strip()
        comment = self.get_lims_comment(sample_id=sample_obj.internal_id)
        has_comment = bool(comment)

        if "gonorrhoeae" in organism:
            organism = "Neisseria spp."
        elif "Cutibacterium acnes" in organism:
            organism = "Propionibacterium acnes"

        if organism == "VRE":
            reference = sample_obj.organism.reference_genome
            if reference == "NC_017960.1":
                organism = "Enterococcus faecium"
            elif reference == "NC_004668.1":
                organism = "Enterococcus faecalis"
            elif has_comment:
                organism = comment

        return organism

    def get_parameters(self, sample_obj: models.Sample) -> Dict[str, str]:
        """Fill a dict with case config information for one sample """

        sample_id = sample_obj.internal_id
        method_library_prep = self.lims.get_prep_method(sample_id)
        if method_library_prep:
            method_library_prep, _ = method_library_prep.split(" ", 1)
        method_sequencing = self.lims.get_sequencing_method(sample_id)
        if method_sequencing:
            method_sequencing, _ = method_sequencing.split(" ", 1)
        priority = "research" if sample_obj.priority == 0 else "standard"

        return {
            "CG_ID_project": self.get_project(sample_id),
            "Customer_ID_project": sample_obj.ticket_number,
            "CG_ID_sample": sample_obj.internal_id,
            "Customer_ID_sample": sample_obj.name,
            "organism": self.get_organism(sample_obj),
            "priority": priority,
            "reference": sample_obj.reference_genome,
            "Customer_ID": sample_obj.customer.internal_id,
            "application_tag": sample_obj.application_version.application.tag,
            "date_arrival": str(sample_obj.received_at or datetime.min),
            "date_sequencing": str(sample_obj.sequenced_at or datetime.min),
            "date_libprep": str(sample_obj.prepared_at or datetime.min),
            "method_libprep": method_library_prep or "Not in LIMS",
            "method_sequencing": method_sequencing or "Not in LIMS",
        }

    def get_project(self, sample_id: str) -> str:
        """Get LIMS project for a sample"""
        return self.lims.get_sample_project(sample_id)

    def get_deliverables_to_store(self) -> list:
        """Retrieve a list of microbial deliverables files for orders where analysis finished
        successfully, and are ready to be stored in Housekeeper"""

        return self.db.cases_to_store(pipeline=Pipeline.MICROSALT)

    def set_statusdb_action(self, case_id: str, action: Optional[str]) -> None:
        """Sets action on case based on ticket number"""
        if action in [None, *CASE_ACTIONS]:
            case_object: models.Family = self.db.family(case_id)
            case_object.action = action
            self.db.commit()
            LOG.info("Action %s set for case %s", action, case_id)
            return
        LOG.warning(
            f"Action '{action}' not permitted by StatusDB and will not be set for case {case_id}"
        )

    def resolve_case_sample_id(
        self, sample: bool, ticket: bool, unique_id: Any
    ) -> (str, Optional[str]):
        """Resolve case_id and sample_id w based on input arguments. """
        if ticket and sample:
            LOG.error("Flags -t and -s are mutually exclusive!")
            raise click.Abort

        if ticket:
            case_id, sample_id = self.get_case_id_from_ticket(unique_id)

        elif sample:
            case_id, sample_id = self.get_case_id_from_sample(unique_id)

        else:
            case_id, sample_id = self.get_case_id_from_case(unique_id)

        return case_id, sample_id

    def get_case_id_from_ticket(self, unique_id: str) -> (str, None):
        """If ticked is provided as argument, finds the corresponding case_id and returns it.
        Since sample_id is not specified, nothing is returned as sample_id"""
        case_obj: models.Family = self.db.find_family_by_name(unique_id)
        if not case_obj:
            LOG.error("No case found for ticket number:  %s", unique_id)
            raise click.Abort
        case_id = case_obj.internal_id
        return case_id, None

    def get_case_id_from_sample(self, unique_id: str) -> (str, str):
        """If sample is specified, finds the corresponding case_id to which this sample belongs.
        The case_id is to be used for identifying the appropriate path to link fastq files and store the analysis output
        """
        sample_obj: models.Sample = (
            self.db.query(models.Sample).filter(models.Sample.internal_id == unique_id).first()
        )
        if not sample_obj:
            LOG.error("No sample found with id: %s", unique_id)
            raise click.Abort
        case_id = sample_obj.links[0].family.internal_id
        sample_id = sample_obj.internal_id
        return case_id, sample_id

    def get_case_id_from_case(self, unique_id: str) -> (str, None):
        """If case_id is specified, validates the presence of case_id in database and returns it"""
        case_obj: models.Family = self.db.family(unique_id)
        if not case_obj:
            LOG.error("No case found with the id:  %s", unique_id)
            raise click.Abort
        case_id = case_obj.internal_id
        return case_id, None

    def store_microbial_analysis_housekeeper(self, case_id: str) -> None:
        """Gather information from microSALT analysis to store."""

        deliverables_path = self.get_deliverables_file_path(case_id=case_id)
        if not deliverables_path.exists():
            LOG.warning(
                "Deliverables file not found for %s, analysis may not be finished yet", case_id
            )
            raise click.Abort

        analysis_date = self.get_date_from_deliverables_path(deliverables_path=deliverables_path)

        bundle_data = self.hermes_api.create_housekeeper_bundle(
            deliverables=deliverables_path,
            pipeline="microsalt",
            created=analysis_date,
            analysis_type=None,
            bundle_name=case_id,
        ).dict()
        bundle_data["name"] = case_id

        bundle_result = self.hk.add_bundle(bundle_data=bundle_data)
        if not bundle_result:
            raise BundleAlreadyAddedError("Bundle already added to Housekeeper!")
        bundle_object, bundle_version = bundle_result
        self.hk.include(bundle_version)
        self.hk.add_commit(bundle_object, bundle_version)
        LOG.info(
            f"Analysis successfully stored in Housekeeper: {case_id} : {bundle_version.created_at}"
        )

    def store_microbial_analysis_statusdb(self, case_id: str) -> None:
        """Creates an analysis object in StatusDB"""
        deliverables_path = self.get_deliverables_file_path(case_id=case_id)
        analysis_date = self.get_date_from_deliverables_path(deliverables_path=deliverables_path)
        case_obj: models.Family = self.db.family(case_id)

        new_analysis: models.Analysis = self.db.add_analysis(
            pipeline=Pipeline.MICROSALT,
            version=self.get_microsalt_version(),
            started_at=analysis_date,
            completed_at=datetime.now(),
            primary=(len(case_obj.analyses) == 0),
        )
        new_analysis.family = case_obj
        self.db.add_commit(new_analysis)
        LOG.info(f"Analysis successfully stored in StatusDB: {case_id} : {analysis_date}")

    @staticmethod
    def get_date_from_deliverables_path(deliverables_path: Path) -> datetime.date:
        """ Get date from deliverables path using date created metadata """
        return datetime.fromtimestamp(int(os.path.getctime(deliverables_path)))

    def get_priority(self, case_id: str) -> str:
        """Returns priority for the case in clinical-db as text"""
        case_object = self.db.family(case_id)
        if case_object.high_priority:
            return "high"
        if case_object.low_priority:
            return "low"
        return "normal"

    def submit_trailblazer_analysis(self, case_id: str) -> None:
        self.trailblazer_api.mark_analyses_deleted(case_id=case_id)
        self.trailblazer_api.add_pending_analysis(
            case_id=case_id,
            email=environ_email(),
            type="other",
            out_dir=self.get_deliverables_file_path(case_id).parent.as_posix(),
            config_path=self.get_trailblazer_config_path(case_id=case_id).as_posix(),
            priority=self.get_priority(case_id),
            data_analysis=Pipeline.MICROSALT,
        )

    def get_microsalt_version(self) -> str:
        try:
            self.process.run_command(["--version"])
            return list(self.process.stdout_lines())[0].split()[-1]
        except CalledProcessError:
            LOG.warning("Could not retrieve microsalt version!")
            return "0.0.0"



from pathlib import Path
import logging
import csv
from subprocess import CalledProcessError
from typing import Any, Optional, Tuple

import shutil
import datetime as dt

from alchy import Query

from cg.apps.hermes.hermes_api import HermesApi
from cg.exc import BundleAlreadyAddedError, CgError
from cg.utils import Process
from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.apps.tb import TrailblazerAPI
from cg.apps.lims import LimsAPI
from cg.store import Store, models
import os
from cg.constants import Pipeline, CASE_ACTIONS
from housekeeper.store.models import Bundle, Version

LOG = logging.getLogger(__name__)


class FluffyAnalysisAPI:
    def __init__(
        self,
        housekeeper_api: HousekeeperAPI,
        trailblazer_api: TrailblazerAPI,
        hermes_api: HermesApi,
        lims_api: LimsAPI,
        status_db: Store,
        config: dict,
    ):
        self.housekeeper_api = housekeeper_api
        self.trailblazer_api = trailblazer_api
        self.status_db = status_db
        self.lims_api = lims_api
        self.hermes_api = hermes_api
        self.root_dir = Path(config["root_dir"])
        self.process = Process(binary=config["binary_path"])
        self.fluffy_config = Path(config["config_path"])

    def get_samplesheet_path(self, case_id: str) -> Path:
        """
        Location in case folder where samplesheet is expected to be stored. Samplesheet is used as a config
        required to run Fluffy
        """
        return Path(self.root_dir, case_id, "SampleSheet.csv")

    def get_workdir_path(self, case_id: str) -> Path:
        """
        Location in case folder where all sub-folders for each sample containing fastq files are to e saved
        """
        return Path(self.root_dir, case_id, "fastq")

    def get_fastq_path(self, case_id: str, sample_id: str) -> Path:
        """
        Location in working directory to which fastq files are saved for each sample separately
        """
        return Path(self.get_workdir_path(case_id), sample_id)

    def get_output_path(self, case_id: str) -> Path:
        """
        Location in case directory where Fluffy outputs will be stored
        """
        return Path(self.root_dir, case_id, "output")

    def get_deliverables_path(self, case_id: str) -> Path:
        """
        Location in working directory where deliverables file will be stored upon completion of analysis.
        Deliverables file is used to communicate paths and tag definitions for files in a finished analysis
        """
        return Path(self.get_output_path(case_id), "deliverables.yaml")

    def get_slurm_job_ids_path(self, case_id: str) -> Path:
        """
        Location in working directory where SLURM job id file is to be stored.
        This file contains SLURM ID of jobs associated with current analysis ,
        is used as a config to be submitted to Trailblazer and track job progress in SLURM
        """
        return Path(self.get_output_path(case_id), "sacct", "submitted_jobs.yaml")

    def get_priority(self, case_id: str) -> str:
        """Returns priority for the case in clinical-db as text"""
        case_obj: models.Family = self.status_db.family(case_id)
        if case_obj:
            if case_obj.high_priority:
                return "high"
            if case_obj.low_priority:
                return "low"
        return "normal"

    def verify_case_id_in_database(self, case_id: str) -> None:
        """
        Passes silently if case exists in StatusDB, raises error if case is missing
        """
        case_obj = self.status_db.family(case_id)
        if not case_obj:
            LOG.error("Case %s could not be found in StatusDB!", case_id)
            raise CgError

    def get_sample_name_from_lims_id(self, lims_id: str) -> str:
        """
        Retrieve sample name provided by customer for specific sample
        """
        LOG.debug("Setting sample name for %s", lims_id)
        sample_obj = self.status_db.sample(lims_id)
        return sample_obj.name

    def link_fastq_files(self, case_id: str, dry_run: bool) -> None:
        """
        Links fastq files from Housekeeper to case working directory
        """
        case_obj: models.Family = self.status_db.family(case_id)
        workdir_path = self.get_workdir_path(case_id=case_id)
        if workdir_path.exists() and not dry_run:
            LOG.info("Fastq directory exists, removing and re-linking files!")
            shutil.rmtree(workdir_path, ignore_errors=True)
        workdir_path.mkdir(parents=True, exist_ok=True)
        for familysample in case_obj.links:
            sample_id = familysample.sample.internal_id
            files: Query = self.housekeeper_api.files(bundle=sample_id, tags=["fastq"])
            sample_path: Path = self.get_fastq_path(case_id=case_id, sample_id=sample_id)
            for file in files:
                if not dry_run:
                    Path.mkdir(sample_path, exist_ok=True, parents=True)
                    Path(sample_path / Path(file.full_path).name).symlink_to(file.full_path)
                LOG.info(f"Linking {file.full_path} to {sample_path / Path(file.full_path).name}")

    def get_concentrations_from_lims(self, sample_id: str) -> str:
        """Get sample concentration from LIMS"""
        return self.lims_api.get_sample_attribute(lims_id=sample_id, key="concentration_sample")

    def add_concentrations_to_samplesheet(
        self, samplesheet_housekeeper_path: Path, samplesheet_workdir_path: Path
    ) -> None:
        """
        Reads the fluffy samplesheet *.csv file as found in Housekeeper.
        Edits column 'SampleName' to include customer name for sample.
        Appends a column 'Library_nM' at the end of the csv with concentration values extracted from LIMS
        """
        csv_reader = csv.reader(open(samplesheet_housekeeper_path, "r"))
        csv_writer = csv.writer(open(samplesheet_workdir_path, "w"))

        sample_name_index = None
        sample_id_index = None
        for row in csv_reader:
            if "SampleID" in row and not sample_id_index:
                sample_name_index = row.index("SampleName")
                sample_id_index = row.index("SampleID")
                row.append("Library_nM")
                csv_writer.writerow(row)
                continue
            if sample_id_index:
                sample_id = row[sample_id_index]
                row[sample_name_index] = self.get_sample_name_from_lims_id(lims_id=sample_id)
                row.append(str(self.get_concentrations_from_lims(sample_id=sample_id)))
                csv_writer.writerow(row)

    def get_samplesheet_housekeeper_path(self, flowcell_name: str) -> Path:
        """
        Returns the path to original samplesheet file that is added to Housekeeper
        """
        samplesheet_query: list = self.housekeeper_api.files(
            bundle=flowcell_name, tags=["samplesheet"]
        ).all()
        if not samplesheet_query:
            LOG.error(
                "Samplesheet file for flowcell %s could not be found in Housekeeper!", flowcell_name
            )
            raise CgError
        return Path(samplesheet_query[0].full_path)

    def make_samplesheet(self, case_id: str, dry_run: bool) -> None:
        """
        Create SampleSheet.csv file in working directory and add desired values to the file
        """
        case_obj: models.Family = self.status_db.family(case_id)
        flowcell_name: str = case_obj.links[0].sample.flowcells[0].name
        samplesheet_housekeeper_path = self.get_samplesheet_housekeeper_path(
            flowcell_name=flowcell_name
        )
        samplesheet_workdir_path = Path(self.get_samplesheet_path(case_id=case_id))
        LOG.info(
            "Writing modified csv from %s to %s",
            samplesheet_housekeeper_path,
            samplesheet_workdir_path,
        )
        if not dry_run:
            Path(self.root_dir, case_id).mkdir(parents=True, exist_ok=True)
            self.add_concentrations_to_samplesheet(
                samplesheet_housekeeper_path=samplesheet_housekeeper_path,
                samplesheet_workdir_path=samplesheet_workdir_path,
            )

    def run_fluffy(self, case_id: str, dry_run: bool) -> None:
        """
        Call fluffy with the configured command-line arguments
        """
        output_path: Path = self.get_output_path(case_id=case_id)
        if output_path.exists():
            LOG.info("Old working directory found, cleaning!")
            if not dry_run:
                shutil.rmtree(output_path, ignore_errors=True)
        command_args = [
            "--config",
            self.fluffy_config.as_posix(),
            "--sample",
            self.get_samplesheet_path(case_id=case_id).as_posix(),
            "--project",
            self.get_workdir_path(case_id=case_id).as_posix(),
            "--out",
            self.get_output_path(case_id=case_id).as_posix(),
            "analyse",
        ]
        self.process.run_command(command_args, dry_run=dry_run)

    def upload_bundle_housekeeper(self, case_id: str) -> None:
        """
        Save the completed analysis bundle in Housekeeper
        """
        deliverables_path = self.get_deliverables_path(case_id=case_id)
        if not deliverables_path.exists():
            LOG.error("Deliverables file not found for case %s, aborting!", case_id)
            raise CgError
        analysis_date = self.get_date_from_file_path(file_path=deliverables_path)
        bundle_data = self.hermes_api.create_housekeeper_bundle(
            deliverables=deliverables_path,
            pipeline="fluffy",
            created=analysis_date,
            analysis_type=None,
            bundle_name=case_id,
        ).dict()
        bundle_data["name"] = case_id
        bundle_result: Tuple[Bundle, Version] = self.housekeeper_api.add_bundle(
            bundle_data=bundle_data
        )
        if not bundle_result:
            raise BundleAlreadyAddedError("Bundle already added to Housekeeper!")
        bundle_object, bundle_version = bundle_result
        self.housekeeper_api.include(bundle_version)
        self.housekeeper_api.add_commit(bundle_object, bundle_version)
        LOG.info(
            f"Analysis successfully stored in Housekeeper: {case_id} : {bundle_version.created_at}"
        )

    def upload_bundle_statusdb(self, case_id: str) -> None:
        """
        Create completed analysis object in StatusDB
        """
        case_obj: models.Family = self.status_db.family(case_id)
        analysis_date = self.get_date_from_file_path(
            file_path=self.get_deliverables_path(case_id=case_id)
        )
        new_analysis: models.Analysis = self.status_db.add_analysis(
            pipeline=Pipeline.FLUFFY,
            started_at=self.get_date_from_file_path(
                file_path=self.get_samplesheet_path(case_id=case_id)
            ),
            version=self.get_pipeline_version(),
            completed_at=self.get_date_from_file_path(
                file_path=self.get_deliverables_path(case_id=case_id)
            ),
            primary=(len(case_obj.analyses) == 0),
        )
        new_analysis.family = case_obj
        self.status_db.add_commit(new_analysis)
        LOG.info(f"Analysis successfully stored in StatusDB: {case_id} : {analysis_date}")

    @staticmethod
    def get_date_from_file_path(file_path: Path) -> dt.datetime.date:
        """
        Get date from deliverables path using date created metadata.
        """
        return dt.datetime.fromtimestamp(int(os.path.getctime(file_path)))

    def get_cases_to_store(self) -> list:
        """
        Get a list of cases with action 'running' and existing deliverables file.
        """
        return [
            case_obj
            for case_obj in self.status_db.cases_to_store(pipeline=Pipeline.FLUFFY)
            if self.get_deliverables_path(case_id=case_obj.internal_id).exists()
        ]

    def get_pipeline_version(self) -> str:
        """
        Calls the pipeline to get the pipeline version number. If fails, returns a placeholder value instead.
        """
        try:
            self.process.run_command(["--version"])
            return list(self.process.stdout_lines())[0].split()[-1]
        except (Exception, CalledProcessError):
            LOG.warning("Could not retrieve fluffy version!")
            return "0.0.0"

    def set_statusdb_action(self, case_id: str, action: Optional[str]) -> None:
        """
        Set one of the allowed actions on a case in StatusDB.
        """
        if action in [None, *CASE_ACTIONS]:
            case_obj: models.Family = self.status_db.family(case_id)
            case_obj.action = action
            self.status_db.commit()
            LOG.info("Action %s set for case %s", action, case_id)
            return
        LOG.warning(
            f"Action '{action}' not permitted by StatusDB and will not be set for case {case_id}"
        )



"""Module with Convenience methods for cg samples"""
from cg.constants import Pipeline
from cg.store import Store


class SampleHelper:
    """API with Convenience methods for cg samples"""

    def __init__(self, store: Store):
        self._store = store

    def is_ready_made_sample(self, internal_id):
        """Check if sample is a rml sample"""
        return self.get_prep_category(internal_id) == "rml"

    def is_externally_sequenced(self, internal_id):
        """Check if sample is externally sequenced"""
        return self._store.sample(internal_id).application_version.application.is_external

    def get_prep_category(self, internal_id):
        """Get the prep category of the application of the sample"""
        return self._store.sample(internal_id).application_version.application.prep_category

    def is_analysis_sample(self, internal_id):
        """Check if this sample is destined for an analysis"""
        return (
            not self._store.family(internal_id).data_analysis
            or str(Pipeline.FASTQ) not in self._store.family(internal_id).data_analysis
        )



import logging

import click

from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.apps.lims import LimsAPI
from cg.apps.stats import StatsAPI
from cg.meta import transfer as transfer_app
from cg.store import Store

log = logging.getLogger(__name__)


@click.group()
@click.pass_context
def transfer(context):
    """Transfer results to the status interface."""
    context.obj["status_db"] = Store(context.obj["database"])


@transfer.command()
@click.argument("flowcell_name")
@click.pass_context
def flowcell(context, flowcell_name):
    """Populate results from a flowcell."""
    stats_api = StatsAPI(context.obj)
    hk_api = HousekeeperAPI(context.obj)
    transfer_api = transfer_app.TransferFlowcell(context.obj["status_db"], stats_api, hk_api)
    new_record = transfer_api.transfer(flowcell_name)
    context.obj["status_db"].add_commit(new_record)
    click.echo(click.style(f"flowcell added: {new_record}", fg="green"))


@transfer.command()
@click.option(
    "-s", "--status", type=click.Choice(["received", "prepared", "delivered"]), default="received"
)
@click.option(
    "-i", "--include", type=click.Choice(["unset", "not-invoiced", "all"]), default="unset"
)
@click.option("--sample-id", help="Lims Submitted Sample id. use together with status.")
@click.pass_context
def lims(context, status, include, sample_id):
    """Check if samples have been updated in LIMS."""
    lims_api = LimsAPI(context.obj)
    transfer_api = transfer_app.TransferLims(context.obj["status_db"], lims_api)
    transfer_api.transfer_samples(transfer_app.SampleState[status.upper()], include, sample_id)


@transfer.command()
@click.option("-s", "--status", type=click.Choice(["received", "delivered"]), default="delivered")
@click.pass_context
def pools(context, status):
    """
    Update pools with received_at or delivered_at dates from LIMS. Defaults to delivered if no
    option is provided.
    """
    lims_api = LimsAPI(context.obj)
    transfer_api = transfer_app.TransferLims(context.obj["status_db"], lims_api)
    transfer_api.transfer_pools(transfer_app.PoolState[status.upper()])



import datetime as dt

import pytest

from cg.apps.lims import LimsAPI
from cg.apps.stats import StatsAPI
from cg.meta.transfer import TransferLims
from cg.meta.transfer.flowcell import TransferFlowcell


@pytest.fixture(name="data")
def fixture_data():
    return {
        "samples": [
            {
                "name": "ADM1136A3",
                "index": "ACGTACAT",
                "flowcell": "HJKMYBCXX",
                "type": "hiseqx",
            }
        ]
    }


@pytest.fixture(scope="function")
def store_stats():
    """Setup base CGStats store."""
    _store = StatsAPI({"cgstats": {"database": "sqlite://", "root": "tests/fixtures/DEMUX"}})
    _store.create_all()
    yield _store
    _store.drop_all()


@pytest.fixture(scope="function")
def base_store_stats(store_stats, data):
    """Setup CGStats store with sample data."""
    demuxes = {}
    for sample_data in data["samples"]:
        project = store_stats.Project(projectname="test", time=dt.datetime.now())
        sample = store_stats.Sample(
            samplename=sample_data["name"],
            barcode=sample_data["index"],
            limsid=sample_data["name"],
        )
        sample.project = project
        unaligned = store_stats.Unaligned(readcounts=300000000, q30_bases_pct=85)
        unaligned.sample = sample

        if sample_data["flowcell"] in demuxes:
            demux = demuxes[sample_data["flowcell"]]
        else:
            flowcell = store_stats.Flowcell(
                flowcellname=sample_data["flowcell"],
                flowcell_pos="A",
                hiseqtype=sample_data["type"],
                time=dt.datetime.now(),
            )
            supportparams = store_stats.Supportparams(document_path="NA", idstring="NA")
            datasource = store_stats.Datasource(document_path="NA", document_type="html")
            datasource.supportparams = supportparams
            demux = store_stats.Demux()
            demux.flowcell = flowcell
            demux.datasource = datasource
            demuxes[sample_data["flowcell"]] = demux

        unaligned.demux = demux
        store_stats.add(unaligned)
    store_stats.commit()
    yield store_stats


@pytest.fixture(scope="function")
def flowcell_store(base_store, data):
    """Setup store with sample data for testing flowcell transfer."""
    for sample_data in data["samples"]:
        customer_obj = base_store.customers().first()
        application_version = base_store.application("WGTPCFC030").versions[0]
        sample = base_store.add_sample(name="NA", sex="male", internal_id=sample_data["name"])
        sample.customer = customer_obj
        sample.application_version = application_version
        sample.received_at = dt.datetime.now()
        base_store.add(sample)
    base_store.commit()
    yield base_store


@pytest.fixture(scope="function")
def transfer_flowcell_api(flowcell_store, housekeeper_api, base_store_stats):
    """Setup flowcell transfer API."""
    transfer_api = TransferFlowcell(flowcell_store, base_store_stats, housekeeper_api)
    yield transfer_api


@pytest.fixture(scope="function")
def transfer_lims_api(sample_store):
    """Setup flowcell transfer API."""
    transfer_api = TransferLims(sample_store, MockLims(config=""))
    yield transfer_api


class MockLims(LimsAPI):
    def __init__(self, config):
        pass

    _received_at = None
    _delivered_at = None
    _prepared_date = None
    _samples = []

    def get_received_date(self, lims_id: str):

        received_date = None
        for sample in self._samples:
            if sample.internal_id == lims_id:
                received_date = sample.received_at
        return received_date

    def mock_set_samples(self, samples):
        self._samples = samples


@pytest.fixture(scope="function")
def lims_api():

    _lims_api = MockLims(config="")
    return _lims_api



"""This script tests the cli mip store functions"""
import logging

from cg.apps.tb.models import TrailblazerAnalysis
from cg.cli.workflow.mip.store import analysis, completed
from cg.constants import EXIT_FAIL, EXIT_SUCCESS
from click.testing import CliRunner


def test_store_no_config(cli_runner: CliRunner, mip_store_context: dict, caplog):
    """Test the function to check for config files"""

    caplog.set_level(logging.INFO)
    # WHEN we run store analysis on no config
    result = cli_runner.invoke(analysis, obj=mip_store_context)
    # THEN We should be informed to enter a config file
    assert "Provide a config file." in caplog.text
    # THEN the exit code should be EXIT_FAIL
    assert result.exit_code == EXIT_FAIL


def test_store_analysis_exception(
    cli_runner: CliRunner, mip_store_context: dict, mip_configs: dict, caplog
):
    """Test that the analysis function enters the exception clause"""
    with caplog.at_level("ERROR", "INFO"):
        # WHEN we run store on a case without deliverables
        result = cli_runner.invoke(
            analysis, [str(mip_configs["purplesnail"])], obj=mip_store_context
        )
        # THEN we should be informed that mandatory files are missing
        assert "Mandatory files are missing" in caplog.text
        # THEN we should not include any files in housekeeper
        assert "included files in Housekeeper" not in caplog.text
        # THEN the exit code should be EXIT_FAIL
        assert result.exit_code == EXIT_FAIL


def test_store_analysis(
    cli_runner: CliRunner,
    mip_configs: dict,
    mip_store_context: dict,
    mip_qc_sample_info: dict,
    caplog,
):
    """Test if store completed stores a completed sample"""

    with caplog.at_level("INFO"):
        # WHEN we run store on a config
        result = cli_runner.invoke(analysis, [str(mip_configs["yellowhog"])], obj=mip_store_context)
        # THEN we should store the files in housekeeper
        assert "new bundle added: yellowhog" in caplog.text
        assert "Included files in Housekeeper" in caplog.text
        # THEN the exit code should be EXIT_SUCCESS
        assert result.exit_code == EXIT_SUCCESS


def test_store_completed_good_cases(
    cli_runner: CliRunner, mip_store_context: dict, mip_case_ids: dict, mip_configs, helpers, caplog
):
    """Test if store completed stores function"""

    with caplog.at_level("INFO"):
        trailblazer_api = mip_store_context["trailblazer_api"]
        analyses = [
            {
                "id": 1,
                "family": "yellowhog",
                "config_path": mip_configs["yellowhog"].as_posix(),
                "status": "completed",
            },
            {
                "id": 2,
                "family": "bluezebra",
                "config_path": mip_configs["bluezebra"].as_posix(),
                "status": "completed",
            },
            {
                "id": 3,
                "family": "purplesnail",
                "config_path": mip_configs["purplesnail"].as_posix(),
                "status": "completed",
            },
        ]
        for analysis in analyses:
            trailblazer_api.ensure_get_latest_analysis_response(analysis_dict=analysis)

        status_db = mip_store_context["status_db"]
        for case_id in ["yellowhog", "bluezebra", "purplesnail"]:
            case_obj = status_db.family(case_id)
            if case_obj:
                case_obj.action = "running"
                status_db.commit()
            else:
                helpers.add_case(store=status_db, internal_id=case_id, action="running")

        # WHEN we run store all completed cases
        result = cli_runner.invoke(completed, obj=mip_store_context)
        # THEN some cases should be added and some should fail
        assert "new bundle added: yellowhog" in caplog.text
        assert "Case storage failed: purplesnail" in caplog.text
        assert "new bundle added: bluezebra" in caplog.text
        assert "Included files in Housekeeper" in caplog.text
        # THEN the command should have an EXIT_FAIL code
        assert result.exit_code == EXIT_FAIL



from cg.apps.lims import orderform
from cg.constants import DataDelivery, Pipeline


def test_parsing_rml_orderform(rml_orderform: str):
    # GIVEN a path to a RML orderform with 2 sample in a pool
    # WHEN parsing the file
    data = orderform.parse_orderform(rml_orderform)

    # THEN it should determine the type of project and customer
    assert data["project_type"] == "rml"
    assert data["customer"] == "cust000"
    # ... and find all samples
    assert len(data["items"]) == 26

    # ... and collect relevant sample data
    sample_data = data["items"][0]
    assert sample_data["name"] == "sample1"
    assert sample_data["pool"] == "pool1"
    assert sample_data["application"] == "RMLP10R300"
    assert sample_data["data_analysis"].lower() == str(Pipeline.FLUFFY)
    assert sample_data["volume"] == "1"
    assert sample_data["concentration"] == "2"
    assert sample_data["index"] == "IDT DupSeq 10 bp Set B"
    assert sample_data["index_number"] == "1"

    assert sample_data["container_name"] is None
    assert sample_data["rml_plate_name"] == "plate"
    assert sample_data["well_position"] is None
    assert sample_data["well_position_rml"] == "A:1"

    assert sample_data["reagent_label"] == "A01 IDT_10nt_541 (ATTCCACACT-AACAAGACCA)"

    assert sample_data["custom_index"] == "GATACA"

    assert sample_data["comment"] == "comment"
    assert sample_data["concentration_sample"] == "3"


def test_parsing_fastq_orderform(fastq_orderform):

    # GIVEN a FASTQ orderform with 2 samples, one normal, one tumour
    # WHEN parsing the file
    data = orderform.parse_orderform(fastq_orderform)

    # THEN it should determine the project type
    assert data["project_type"] == str(Pipeline.FASTQ)
    # ... and find all samples
    assert len(data["items"]) == 36

    # ... and collect relevant sample info
    normal_sample = data["items"][1]
    tumour_sample = data["items"][0]
    assert normal_sample["name"] == "s2"
    assert normal_sample["container"] == "Tube"
    assert normal_sample["data_analysis"] == "No analysis"
    assert normal_sample["data_delivery"].lower() == str(DataDelivery.FASTQ)
    assert normal_sample["application"] == "WGTPCFC030"
    assert normal_sample["sex"] == "female"
    assert normal_sample["case"] == "c1"
    assert data["customer"] == "cust000"
    assert normal_sample["require_qcok"] is False
    assert normal_sample["source"] == "bone marrow"
    assert normal_sample["priority"] == "research"

    assert normal_sample["container_name"] == ""
    assert normal_sample["well_position"] == ""

    assert normal_sample["tumour"] is False

    assert tumour_sample["tumour"] is True
    assert tumour_sample["source"] == "tissue (FFPE)"

    assert tumour_sample["quantity"] == "4"
    assert tumour_sample["comment"] == "other Elution buffer"


def test_parsing_mip_orderform(mip_orderform):

    # GIVEN an order form for a Scout order with 4 samples, 1 trio, in a plate
    # WHEN parsing the order form
    data = orderform.parse_orderform(mip_orderform)

    # THEN it should detect the type of project
    assert data["project_type"] == str(Pipeline.MIP_DNA)
    assert data["customer"] == "cust000"

    # ... and it should find and group all samples in families
    assert len(data["items"]) == 36

    # ... and collect relevant data about the families
    trio_family = data["items"][0]
    assert len(trio_family["samples"]) == 3
    assert trio_family["name"] == "c1"
    assert trio_family["priority"] == "research"
    assert set(trio_family["panels"]) == set(["AD-HSP", "Ataxi", "ATX"])
    assert trio_family["require_qcok"] is True

    # ... and collect relevant info about the samples
    proband_sample = trio_family["samples"][0]
    assert proband_sample["name"] == "s1"
    assert proband_sample["container"] == "96 well plate"
    assert proband_sample["data_analysis"] == "MIP DNA"
    assert proband_sample["data_delivery"].lower() == str(DataDelivery.SCOUT)
    assert proband_sample["application"] == "PANKTTR010"
    assert proband_sample["sex"] == "male"
    assert proband_sample["source"] == "tissue (FFPE)"
    assert proband_sample["tumour"] is True
    assert proband_sample["container_name"] == "plate1"
    assert proband_sample["well_position"] == "A:1"
    assert proband_sample["status"] == "affected"
    assert proband_sample["mother"] == "s2"
    assert proband_sample["father"] == "s3"
    assert proband_sample["quantity"] == "4"
    assert proband_sample["comment"] == "other Elution buffer"


def test_parsing_external_orderform(external_orderform):

    # GIVEN an orderform for two external samples, one WES, one WGS
    # WHEN parsing the file
    data = orderform.parse_orderform(external_orderform)

    # THEN it should detect the project type
    assert data["project_type"] == "external"
    assert data["customer"] == "cust002"

    # ... and find all families (1) and samples (2)
    assert len(data["items"]) == 4
    family = data["items"][0]
    assert family["name"] == "fam2"
    assert family["priority"] == "standard"
    assert set(family["panels"]) == set(["CILM", "CTD"])
    # todo: assert set(case['additional_gene_list']) == set(['16PDEL'])

    # ... and collect info about the samples
    wes_sample = family["samples"][0]
    wgs_sample = family["samples"][1]

    assert wes_sample["capture_kit"] == "Agilent Sureselect V5"
    assert wes_sample["application"] == "EXXCUSR000"
    assert wes_sample["data_analysis"] == "scout"
    assert wes_sample["sex"] == "male"
    # case name on case
    # priority on case
    # customer on order (data)
    assert wes_sample["source"] == "blood"

    # panels on case
    # additional gene list on case
    assert wes_sample["status"] == "affected"

    assert wes_sample["mother"] == "mother"
    assert wes_sample["father"] == "father"
    # todo: assert wes_sample['other'] == 'other', check if removed in latest OF

    # todo: assert wes_sample['tumour'] == 'Tumor'
    # todo: assert wes_sample['gel_picture'] == 'Y'
    # todo: assert wes_sample['extraction_method'] == 'extraction method'
    assert wes_sample["comment"] == "sample comment"

    assert wgs_sample.get("capture_kit") == "Agilent Sureselect CRE"


def test_parsing_metagenome_orderform(metagenome_orderform):

    # GIVEN an orderform for one metagenome sample
    # WHEN parsing the file
    data = orderform.parse_orderform(metagenome_orderform)

    # THEN it should detect the project type
    assert data["project_type"] == "metagenome"
    # ... and find all samples
    assert len(data["items"]) == 19
    # ... and collect relevant sample info
    sample = data["items"][0]

    assert sample["name"] == "sample1"
    assert sample["source"] == "other"
    assert sample["data_analysis"].lower() == str(Pipeline.FASTQ)
    assert sample["application"] == "METPCFR030"
    assert sample["customer"] == "cust000"
    assert sample["require_qcok"] is True
    assert sample["elution_buffer"] == 'Other (specify in "Comments")'
    assert sample["extraction_method"] == "other (specify in comment field)"
    assert sample["container"] == "96 well plate"
    assert sample["priority"] == "research"

    # Required if Plate
    assert sample["container_name"] == "plate1"
    assert sample["well_position"] == "A:1"

    # These fields are not required
    assert sample["concentration_sample"] == "1"
    assert sample["quantity"] == "2"
    assert sample["comment"] == "comment"


def test_parsing_microbial_orderform(microbial_orderform):
    # GIVEN a path to a microbial orderform with 3 samples

    # WHEN parsing the file
    data = orderform.parse_orderform(microbial_orderform)

    # THEN it should determine the type of project and customer
    assert data["project_type"] == str(Pipeline.MICROSALT)
    assert data["customer"] == "cust000"

    # ... and find all samples
    assert len(data["items"]) == 14

    # ... and collect relevant sample data
    sample_data = data["items"][0]

    assert sample_data["name"] == "s1"
    assert sample_data.get("internal_id") is None
    assert sample_data["organism"] == "other"
    assert sample_data["reference_genome"] == "NC_00001"
    assert sample_data["data_analysis"].lower() == str(Pipeline.FASTQ)
    assert sample_data["application"] == "MWRNXTR003"
    # customer on order (data)
    assert sample_data["require_qcok"] is True
    assert sample_data["elution_buffer"] == 'Other (specify in "Comments")'
    assert sample_data["extraction_method"] == "other (specify in comment field)"
    assert sample_data["container"] == "96 well plate"
    assert sample_data.get("priority") in "research"

    assert sample_data["container_name"] == "plate1"
    assert sample_data["well_position"] == "A:1"

    assert sample_data["organism_other"] == "other species"

    assert sample_data["concentration_sample"] == "1"
    assert sample_data["quantity"] == "2"
    assert sample_data["comment"] == "comment"


def test_parsing_balsamic_orderform(balsamic_orderform):

    # GIVEN an order form for a cancer order with 11 samples,
    # WHEN parsing the order form
    data = orderform.parse_orderform(balsamic_orderform)

    # THEN it should detect the type of project
    assert data["project_type"] == str(Pipeline.BALSAMIC)

    # ... and it should find and group all samples in case
    assert len(data["items"]) == 36

    # ... and collect relevant data about the case
    # ... and collect relevant info about the samples

    case = data["items"][0]
    sample = case["samples"][0]
    assert len(case["samples"]) == 2

    # This information is required

    assert sample["name"] == "s1"
    assert sample["container"] == "96 well plate"
    assert sample["data_analysis"].lower() == str(Pipeline.BALSAMIC)
    assert sample["data_delivery"] == str(DataDelivery.ANALYSIS_BAM_FILES)
    assert sample["application"] == "PANKTTR010"
    assert sample["sex"] == "male"
    assert case["name"] == "c1"
    assert data["customer"] == "cust000"
    assert case["require_qcok"] is True
    assert sample["volume"] == "1"
    assert sample["source"] == "tissue (FFPE)"
    assert case["priority"] == "research"

    # Required if Plate
    assert sample["container_name"] == "plate1"
    assert sample["well_position"] == "A:1"

    # This information is required for panel- or exome analysis
    assert sample["elution_buffer"] == 'Other (specify in "Comments")'

    # This information is required for Balsamic analysis (cancer)
    assert sample["tumour"] is True
    assert sample["capture_kit"] == "GMCKsolid"
    assert sample["tumour_purity"] == "5"

    assert sample["formalin_fixation_time"] == "2"
    assert sample["post_formalin_fixation_time"] == "3"
    assert sample["tissue_block_size"] == "small"

    # This information is optional
    assert sample["quantity"] == "4"
    assert sample["comment"] == "other Elution buffer"


def test_parsing_mip_rna_orderform(mip_rna_orderform):

    # GIVEN an order form for a mip balsamic order with 3 samples, 1 trio, in a plate
    # WHEN parsing the order form
    data = orderform.parse_orderform(mip_rna_orderform)

    # THEN it should detect the type of project
    assert data["project_type"] == str(Pipeline.MIP_RNA)
    assert data["customer"] == "cust000"
    # ... and it should find and group all samples in cases
    assert len(data["items"]) == 12
    # ... and collect relevant data about the cases
    first_case = data["items"][0]
    assert len(first_case["samples"]) == 3
    assert first_case["name"] == "c1"
    assert first_case["priority"] == "research"
    assert set(first_case["panels"]) == set()
    assert first_case["require_qcok"] is True
    # ... and collect relevant info about the samples

    first_sample = first_case["samples"][0]
    assert first_sample["name"] == "s1"
    assert first_sample["container"] == "96 well plate"
    assert first_sample["data_analysis"] == "MIP RNA"
    assert first_sample["data_delivery"].lower() == str(DataDelivery.ANALYSIS_FILES)
    assert first_sample["application"] == "RNAPOAR025"
    assert first_sample["sex"] == "male"
    # case-id on the case
    # customer on the order (data)
    # require-qc-ok on the case
    assert first_sample["source"] == "tissue (FFPE)"

    assert first_sample["container_name"] == "plate1"
    assert first_sample["well_position"] == "A:1"

    assert first_sample["tumour"] is True

    assert first_sample["quantity"] == "4"
    assert first_sample["comment"] == "other Elution buffer"

    # required for RNA samples
    assert first_sample["from_sample"] == "s1"
    assert first_sample["time_point"] == "0"
